"""REST API views for the projectroles app"""

import sys

from ipaddress import ip_address, ip_network
from packaging.version import parse as parse_version
from typing import Optional

from django.conf import settings
from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured
from django.db import transaction
from django.http import HttpRequest
from django.utils import timezone

from rest_framework import serializers
from rest_framework.exceptions import (
    APIException,
    NotAcceptable,
    NotFound,
    PermissionDenied,
)
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    RetrieveAPIView,
    UpdateAPIView,
    DestroyAPIView,
)
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import (
    BasePermission,
    AllowAny,
    IsAuthenticated,
)
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.versioning import AcceptHeaderVersioning
from rest_framework.views import APIView

from drf_spectacular.utils import extend_schema, inline_serializer

from projectroles.app_settings import AppSettingAPI
from projectroles.forms import INVITE_EXISTS_MSG
from projectroles.models import (
    Project,
    Role,
    RoleAssignment,
    ProjectInvite,
    RemoteSite,
    AppSetting,
    SODARUser,
    SODAR_CONSTANTS,
    CAT_DELIMITER,
    ROLE_RANKING,
    ROLE_PROJECT_TYPE_ERROR_MSG,
)
from projectroles.plugins import PluginAppSettingDef, PluginAPI
from projectroles.remote_projects import RemoteProjectAPI
from projectroles.serializers import (
    ProjectSerializer,
    RoleAssignmentSerializer,
    ProjectInviteSerializer,
    AppSettingSerializer,
    SODARUserSerializer,
    REMOTE_MODIFY_MSG,
)
from projectroles.utils import get_display_name
from projectroles.views import (
    ProjectAccessMixin,
    ProjectDeleteMixin,
    ProjectDeleteAccessMixin,
    RoleAssignmentDeleteMixin,
    RoleAssignmentOwnerTransferMixin,
    ProjectInviteMixin,
    ProjectModifyPluginViewMixin,
    PROJECT_BLOCK_MSG,
)


app_settings = AppSettingAPI()
plugin_api = PluginAPI()
User = auth.get_user_model()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_ROLE_FINDER = SODAR_CONSTANTS['PROJECT_ROLE_FINDER']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_SCOPE_PROJECT_USER = SODAR_CONSTANTS[
    'APP_SETTING_SCOPE_PROJECT_USER'
]

# API constants for projectroles APIs
PROJECTROLES_API_MEDIA_TYPE = (
    'application/vnd.bihealth.sodar-core.projectroles+json'
)
PROJECTROLES_API_DEFAULT_VERSION = '2.0'
PROJECTROLES_API_ALLOWED_VERSIONS = ['1.0', '1.1', '2.0']
SYNC_API_MEDIA_TYPE = (
    'application/vnd.bihealth.sodar-core.projectroles.sync+json'
)
SYNC_API_DEFAULT_VERSION = '2.0'
SYNC_API_ALLOWED_VERSIONS = ['1.0', '2.0']

# Local constants
APP_NAME = 'projectroles'
INVALID_PROJECT_TYPE_MSG = (
    'Project type "{project_type}" not allowed for this API view'
)
USER_MODIFIABLE_MSG = 'Updating non-user modifiable settings is not allowed'
ANON_ACCESS_MSG = 'Anonymous access not allowed'
NO_VALUE_MSG = 'Value not set in request data'
VIEW_NOT_ACCEPTABLE_VERSION_MSG = (
    'This view is not available in the given API version'
)
USER_LIST_RESTRICT_MSG = (
    'User details access restricted: user does not have contributor access or '
    'above in any category or project '
    '(PROJECTROLES_API_USER_DETAIL_RESTRICT=True)'
)
USER_LIST_INCLUDE_VERSION_MSG = (
    'The include_system_users parameter is not available in API version <1.1'
)
CAT_PUBLIC_STATS_API_MSG = (
    'Setting category_public_stats is only allowed for top level categories'
)
VERSION_1_1 = parse_version('1.1')
VERSION_2_0 = parse_version('2.0')


# Permission / Versioning / Renderer Classes -----------------------------------


class SODARAPIProjectPermission(ProjectAccessMixin, BasePermission):
    """
    Mixin for providing basic project permission checking for API views with a
    single permission_required attribute. Also works with Knox token based
    views.

    This must be used in the ``permission_classes`` attribute in order for token
    authentication to work.

    Requires implementing either ``permission_required`` or
    ``get_permission_required()`` in the view.

    Project type can be restricted to ``PROJECT_TYPE_CATEGORY`` or
    ``PROJECT_TYPE_PROJECT``, as defined in SODAR constants, by setting the
    ``project_type`` attribute in the view.
    """

    def has_permission(self, request, view):
        """
        Override has_permission() for checking auth and project permission
        """
        if (not request.user or request.user.is_anonymous) and not getattr(
            settings, 'PROJECTROLES_ALLOW_ANONYMOUS', False
        ):
            return False
        project = self.get_project(request=request, kwargs=view.kwargs)
        if not project:
            raise NotFound()

        # Restrict project type
        project_type = getattr(view, 'project_type', None)
        p_types = [PROJECT_TYPE_CATEGORY, PROJECT_TYPE_PROJECT]
        if project_type and project_type not in p_types:
            raise ImproperlyConfigured(
                'Invalid value "{}" for project_type, accepted values: '
                '{}'.format(project_type, ', '.join(p_types))
            )
        elif project_type and project_type != project.type:
            raise PermissionDenied(
                INVALID_PROJECT_TYPE_MSG.format(project_type=project.type)
            )

        # Prohibit access if project_access_block is set
        if (
            not request.user.is_superuser
            and project.is_project()
            and app_settings.get(
                APP_NAME, 'project_access_block', project=project
            )
        ):
            raise PermissionDenied(
                PROJECT_BLOCK_MSG.format(
                    project_type=get_display_name(project.type)
                )
            )

        owner_or_delegate = project.is_owner_or_delegate(request.user)
        if not (
            request.user.is_superuser or owner_or_delegate
        ) and app_settings.get(APP_NAME, 'ip_restrict', project):
            for k in (
                'HTTP_X_FORWARDED_FOR',
                'X_FORWARDED_FOR',
                'FORWARDED',
                'REMOTE_ADDR',
            ):
                v = request.META.get(k)
                if v:
                    client_address = ip_address(v.split(',')[0])
                    break
            else:  # Can't fetch client IP address
                return False

            ips = app_settings.get(APP_NAME, 'ip_allow_list', project)
            if not ips:
                return False
            for ip in [s.strip() for s in ips.split(',')]:
                if '/' in ip:
                    if client_address in ip_network(ip):
                        break
                elif client_address == ip_address(ip):
                    break
            else:
                return False

        if not hasattr(view, 'permission_required') and (
            not hasattr(view, 'get_permission_required')
            or not callable(getattr(view, 'get_permission_required', None))
        ):
            raise ImproperlyConfigured(
                '{0} is missing the permission_required attribute. '
                'Define {0}.permission_required, or override '
                '{0}.get_permission_required().'.format(view.__class__.__name__)
            )
        elif hasattr(view, 'permission_required'):
            perm = view.permission_required
        else:
            perm = view.get_permission_required()

        # This may return an iterable, but we are only interested in one perm
        if isinstance(perm, (list, tuple)) and len(perm) > 0:
            # TODO: TBD: Raise exception / log warning if given multiple perms?
            perm = perm[0]
        return request.user.has_perm(perm, project)


# Base API View Mixins ---------------------------------------------------------


class SODARAPIBaseProjectMixin(ProjectAccessMixin):
    """
    API view mixin for the base DRF ``APIView`` class with project permission
    checking, but without serializers and other generic view functionality.

    Project type can be restricted to ``PROJECT_TYPE_CATEGORY`` or
    ``PROJECT_TYPE_PROJECT``, as defined in SODAR constants, by setting the
    ``project_type`` attribute in the view.
    """

    permission_classes = [SODARAPIProjectPermission]
    project_type = None


class APIProjectContextMixin(ProjectAccessMixin):
    """
    Mixin to provide project context and queryset for generic API views. Can
    be used both in SODAR and SODAR Core API base views.

    If your model doesn't have a direct "project" relation, set
    ``queryset_project_field`` in the implementing class to query based on e.g.
    a nested foreignkey relation.
    """

    def get_serializer_context(self, *args, **kwargs):
        context = super().get_serializer_context(*args, **kwargs)
        if sys.argv[1:2] == ['generateschema']:
            return context
        context['project'] = self.get_project(request=context['request'])
        return context

    def get_queryset(self):
        project_field = getattr(self, 'queryset_project_field', 'project')
        return self.__class__.serializer_class.Meta.model.objects.filter(
            **{project_field: self.get_project()}
        )


class SODARAPIGenericProjectMixin(
    APIProjectContextMixin, SODARAPIBaseProjectMixin
):
    """
    API view mixin for generic DRF API views with serializers, SODAR project
    context and permission checkin.

    Unless overriding ``permission_classes`` with their own implementation, the
    user MUST supply a ``permission_required`` attribute.

    Replace ``lookup_url_kwarg`` with your view's url kwarg (SODAR project
    compatible model name in lowercase).

    If the lookup is done via a foreign key, change the ``lookup_field``
    attribute of your class into ``foreignkey__sodar_uuid``, e.g.
    ``project__sodar_uuid`` for lists.

    If your object(s) don't have a direct ``project`` relation, update the
    ``queryset_project_field`` to point to the field, e.g.
    ``someothermodel__project``.
    """

    lookup_field = 'sodar_uuid'  # Use project__sodar_uuid for lists
    lookup_url_kwarg = 'project'  # Replace with relevant model
    queryset_project_field = 'project'  # Replace if no direct project relation


class ProjectQuerysetMixin:
    """
    Mixin for overriding the default queryset with one which allows us to lookup
    a Project object directly.
    """

    def get_queryset(self):
        return Project.objects.all()


class SODARPageNumberPagination(PageNumberPagination):
    """
    Override of PageNumberPagination to provide optional pagination.

    If the "page" query string is not present, results will be provided as a
    full unpaginated list.

    If the "page" query string is included, results will be presented in the
    default ``PageNumberPagination`` dict format.

    See: https://www.django-rest-framework.org/api-guide/pagination/#pagenumberpagination
    """

    def paginate_queryset(self, queryset, request, view=None):
        if not request.query_params.get('page'):
            return None
        return super().paginate_queryset(queryset, request, view)


# SODAR Core Base Views and Mixins ---------------------------------------------


class ProjectrolesAPIVersioningMixin:
    """
    Projectroles API view versioning mixin for overriding media type
    and accepted versions.
    """

    class ProjectrolesAPIVersioning(AcceptHeaderVersioning):
        allowed_versions = PROJECTROLES_API_ALLOWED_VERSIONS
        default_version = PROJECTROLES_API_DEFAULT_VERSION

    class ProjectrolesAPIRenderer(JSONRenderer):
        media_type = PROJECTROLES_API_MEDIA_TYPE

    renderer_classes = [ProjectrolesAPIRenderer]
    versioning_class = ProjectrolesAPIVersioning


class RemoteSyncAPIVersioningMixin:
    """
    Projectroles remote sync API view versioning mixin for overriding media type
    and accepted versions.
    """

    class RemoteSyncAPIRenderer(JSONRenderer):
        media_type = SYNC_API_MEDIA_TYPE

    class RemoteSyncAPIVersioning(AcceptHeaderVersioning):
        allowed_versions = SYNC_API_ALLOWED_VERSIONS
        default_version = SYNC_API_DEFAULT_VERSION

    renderer_classes = [RemoteSyncAPIRenderer]
    versioning_class = RemoteSyncAPIVersioning


class ProjectCreatePermission(ProjectAccessMixin, BasePermission):
    """Permission class specific to Project creation"""

    def has_permission(self, request, view):
        """Override has_permission() to check for project creation permission"""
        parent_uuid = request.data.get('parent')
        parent = (
            Project.objects.filter(sodar_uuid=parent_uuid).first()
            if parent_uuid
            else None
        )
        if (
            parent
            and settings.PROJECTROLES_SITE_MODE == SITE_MODE_TARGET
            and (not settings.PROJECTROLES_TARGET_CREATE or parent.is_remote())
        ):
            return False
        if not parent and not request.user.is_superuser:
            return False
        return request.user.has_perm('projectroles.create_project', parent)


# API Views --------------------------------------------------------------------


class ProjectListAPIView(ProjectrolesAPIVersioningMixin, ListAPIView):
    """
    List all projects and categories for which the requesting user has access.

    Supports optional pagination by providing the ``page`` query string. This
    will return results in the Django Rest Framework ``PageNumberPagination``
    format.

    **URL:** ``/project/api/list``

    **Methods:** ``GET``

    **Parameters:**

    - ``page``: Page number for paginated results (int, optional)

    **Returns:**

    List of ``Project`` objects (see ``ProjectRetrieveAPIView``). For project
    finder role, only lists title and UUID of projects.
    """

    pagination_class = SODARPageNumberPagination
    permission_classes = [IsAuthenticated]
    serializer_class = ProjectSerializer

    def get_queryset(self):
        """
        Override get_queryset() to return categories and projects to which the
        user has access. NOTE: Returns a list, not a QuerySet.

        :return: List of Project objects
        """
        projects_all = Project.objects.all().order_by('full_title')
        if self.request.user.is_superuser:
            qs = projects_all
        else:
            qs = []
            role_cats = []
            projects_local = [
                a.project
                for a in RoleAssignment.objects.filter(user=self.request.user)
            ]
            for p in projects_all:
                local_role = p in projects_local
                if (
                    local_role
                    or p.public_access
                    or any(p.full_title.startswith(c) for c in role_cats)
                ):
                    qs.append(p)
                if local_role and p.is_category():
                    role_cats.append(p.full_title + CAT_DELIMITER)
        return qs


class ProjectRetrieveAPIView(
    ProjectQuerysetMixin,
    ProjectrolesAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    RetrieveAPIView,
):
    """
    Retrieve a project or category by its UUID.

    **URL:** ``/project/api/retrieve/{Project.sodar_uuid}``

    **Methods:** ``GET``

    **Returns:**

    - ``archive``: Project archival status (boolean)
    - ``children``: Category children (list of UUIDs, only returned for categories)
    - ``description``: Project description (string)
    - ``full_title``: Full project title with parent categories (string)
    - ``parent``: Parent category UUID (string or null)
    - ``readme``: Project readme (string, supports markdown)
    - ``public_access``: Read-only access for all users (role name as string or None)
    - ``roles``: Project role assignments (dict, assignment UUID as key)
    - ``sodar_uuid``: Project UUID (string)
    - ``title``: Project title (string)
    - ``type``: Project type (string, options: ``PROJECT`` or ``CATEGORY``)

    **Version Changes:**

    - ``2.0``
        * Replace ``roles`` field user serializer with user UUID
        * Replace ``public_guest_access`` field with ``public_access``
    - ``1.1``
        * Add ``children`` field
    """

    permission_required = 'projectroles.view_project'
    serializer_class = ProjectSerializer


class ProjectCreateAPIView(
    ProjectrolesAPIVersioningMixin, ProjectAccessMixin, CreateAPIView
):
    """
    Create a project or a category.

    If setting public read-only access for users, provide role name ("project
    guest" or "project viewer") or None for no access.

    **URL:** ``/project/api/create``

    **Methods:** ``POST``

    **Parameters:**

    - ``title``: Project title (string)
    - ``type``: Project type (string, options: ``PROJECT`` or ``CATEGORY``)
    - ``parent``: Parent category UUID (string)
    - ``description``: Project description (string, optional)
    - ``readme``: Project readme (string, optional, supports markdown)
    - ``public_access``: Public read-only access for all users (string or None)
    - ``owner``: User UUID of the project owner (string)

    **Version Changes:**

    - ``2.0``
        * Replace ``public_guest_access`` field with ``public_access``
    """

    permission_classes = [ProjectCreatePermission]
    serializer_class = ProjectSerializer


class ProjectUpdateAPIView(
    ProjectQuerysetMixin,
    ProjectrolesAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    UpdateAPIView,
):
    """
    Update the metadata of a project or a category.

    Note that the project owner can not be updated here. Instead, use the
    dedicated API view ``RoleAssignmentOwnerTransferAPIView``.

    The project type can not be updated once a project has been created. The
    parameter is still required for non-partial updates via the ``PUT`` method.

    If setting public read-only access for users, provide role name ("project
    guest" or "project viewer") or None for no access.

    **URL:** ``/project/api/update/{Project.sodar_uuid}``

    **Methods:** ``PUT``, ``PATCH``

    **Parameters:**

    - ``title``: Project title (string)
    - ``type``: Project type (string, can not be modified)
    - ``parent``: Parent category UUID (string)
    - ``description``: Project description (string, optional)
    - ``readme``: Project readme (string, optional, supports markdown)
    - ``public_access``: Public read-only access for all users (string or None)

    - ``2.0``
        * Replace ``public_guest_access`` field with ``public_access``
    """

    permission_required = 'projectroles.update_project'
    serializer_class = ProjectSerializer

    def get_serializer_context(self, *args, **kwargs):
        context = super().get_serializer_context(*args, **kwargs)
        if sys.argv[1:2] == ['generateschema']:
            return context
        project = self.get_project(request=context['request'])
        context['parent'] = (
            str(project.parent.sodar_uuid) if project.parent else None
        )
        return context


class ProjectDestroyAPIView(
    ProjectQuerysetMixin,
    ProjectrolesAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    ProjectDeleteMixin,
    ProjectDeleteAccessMixin,
    DestroyAPIView,
):
    """
    Destroy a project and all associated data.

    Deletion is prohibited if called on a category with children or a project
    with non-revoked remote projects.

    NOTE: This operation can not be undone!

    **URL:** ``/project/api/destroy/{Project.sodar_uuid}``

    **Methods:** ``DELETE``

    **Version Changes:**

    - ``1.1``: Add view
    """

    permission_required = 'projectroles.delete_project'
    serializer_class = ProjectSerializer

    def perform_destroy(self, instance):
        """Override perform_destroy() to handle Project deletion"""
        if parse_version(self.request.version) < VERSION_1_1:
            raise NotAcceptable(VIEW_NOT_ACCEPTABLE_VERSION_MSG)
        access, msg = self.check_delete_permission(instance)
        if not access:
            raise PermissionDenied(msg)
        with transaction.atomic():
            self.handle_delete(instance, self.request)


class RoleAssignmentCreateAPIView(
    ProjectrolesAPIVersioningMixin, SODARAPIGenericProjectMixin, CreateAPIView
):
    """
    Create a role assignment in a project.

    **URL:** ``/project/api/roles/create/{Project.sodar_uuid}``

    **Methods:** ``POST``

    **Parameters:**

    - ``role``: Desired role for user (string, e.g. "project contributor")
    - ``user``: User UUID (string)
    """

    permission_required = 'projectroles.update_project_members'
    serializer_class = RoleAssignmentSerializer


class RoleAssignmentUpdateAPIView(
    ProjectrolesAPIVersioningMixin, SODARAPIGenericProjectMixin, UpdateAPIView
):
    """
    Update the role assignment for a user in a project.

    The user can not be changed in this API view.

    **URL:** ``/project/api/roles/update/{RoleAssignment.sodar_uuid}``

    **Methods:** ``PUT``, ``PATCH``

    **Parameters:**

    - ``role``: Desired role for user (string, e.g. "project contributor")
    - ``user``: User UUID (string)
    """

    lookup_url_kwarg = 'roleassignment'
    permission_required = 'projectroles.update_project_members'
    serializer_class = RoleAssignmentSerializer


class RoleAssignmentDestroyAPIView(
    RoleAssignmentDeleteMixin,
    ProjectrolesAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    DestroyAPIView,
):
    """
    Destroy a role assignment.

    The owner role can not be destroyed using this view.

    **URL:** ``/project/api/roles/destroy/{RoleAssignment.sodar_uuid}``

    **Methods:** ``DELETE``
    """

    lookup_url_kwarg = 'roleassignment'
    permission_required = 'projectroles.update_project_members'
    serializer_class = RoleAssignmentSerializer

    def perform_destroy(self, instance):
        """
        Override perform_destroy() to handle RoleAssignment deletion.
        """
        project = self.get_project()
        # Validation for remote sites and projects
        if project.is_remote():
            raise serializers.ValidationError(REMOTE_MODIFY_MSG)
        # Do not allow editing owner here
        if instance.role.name == PROJECT_ROLE_OWNER:
            raise serializers.ValidationError(
                'Use project updating API to update owner'
            )
        # Check delegate perms
        if (
            instance.role.name == PROJECT_ROLE_DELEGATE
            and not self.request.user.has_perm(
                'projectroles.update_project_delegate', project
            )
        ):
            raise PermissionDenied('User lacks permission to assign delegates')
        self.delete_assignment(role_as=instance, request=self.request)


class RoleAssignmentOwnerTransferAPIView(
    RoleAssignmentOwnerTransferMixin,
    ProjectrolesAPIVersioningMixin,
    SODARAPIBaseProjectMixin,
    APIView,
):
    """
    Trensfer project ownership to another user with a role in the project.
    Reassign a different role to the previous owner.

    The new owner must already have a role assigned in the project.

    NOTE: Barring any inherited role, if no value is given for
    ``old_owner_role``, the old owner's access to the project will be removed.

    **URL:** ``/project/api/roles/owner-transfer/{Project.sodar_uuid}``

    **Methods:** ``POST``

    **Parameters:**

    - ``new_owner``: User name of new owner (string)
    - ``old_owner_role``: Role for old owner (string or None, e.g. "project delegate")

    **Version Changes:**

    - ``1.1``: Allow empty value for ``old_owner_role``
    """

    permission_required = 'projectroles.update_project_owner'
    serializer_class = RoleAssignmentSerializer

    def post(self, request, *args, **kwargs):
        """Handle ownership transfer in a POST request"""
        d_new_owner = request.data.get('new_owner')
        d_old_owner_role = request.data.get('old_owner_role')
        # Validate input
        if not d_new_owner:
            raise serializers.ValidationError(
                'Field "new_owner" must be present'
            )
        # Prevent old_owner_role=None if v1.0
        if (
            parse_version(request.version) < VERSION_1_1
            and not d_old_owner_role
        ):
            raise serializers.ValidationError(
                'Field "old_owner_role" must be present'
            )

        project = self.get_project()
        # Validation for remote sites and projects
        if project.is_remote():
            raise serializers.ValidationError(REMOTE_MODIFY_MSG)

        new_owner = User.objects.filter(username=d_new_owner).first()
        old_owner_role = None
        if d_old_owner_role:
            old_owner_role = Role.objects.filter(name=d_old_owner_role).first()
        old_owner_as = project.get_owner()
        old_owner = old_owner_as.user

        if d_old_owner_role and not old_owner_role:
            raise serializers.ValidationError(
                f'Unknown role "{d_old_owner_role}"'
            )
        if old_owner_role and project.type not in old_owner_role.project_types:
            raise serializers.ValidationError(
                ROLE_PROJECT_TYPE_ERROR_MSG.format(
                    project_type=project.type, role_name=old_owner_role.name
                )
            )
        if not old_owner_as:
            raise serializers.ValidationError('Existing owner role not found')
        if not new_owner:
            raise serializers.ValidationError(f'User "{d_new_owner}" not found')
        if new_owner == old_owner:
            raise serializers.ValidationError('Owner role already set for user')
        if not project.has_role(new_owner):
            raise serializers.ValidationError(
                f'User {new_owner.username} is not a member of the project'
            )
        # Validate existing inherited role for old owner, do not allow demoting
        inh_roles = RoleAssignment.objects.filter(
            user=old_owner, project__in=project.get_parents()
        ).order_by('role__rank')
        if inh_roles and old_owner_role.rank > inh_roles.first().role.rank:
            raise serializers.ValidationError(
                f'User {old_owner.username} has inherited role '
                f'"{inh_roles.first().role.name}", demoting is not allowed'
            )

        try:  # All OK, transfer owner
            self.transfer_owner(
                project,
                new_owner,
                old_owner_as,
                old_owner_role,
                request=request,
            )
        except Exception as ex:
            raise APIException(f'Unable to transfer owner: {ex}')
        return Response(
            {
                'detail': 'Ownership transferred from {} to {} in '
                'project "{}"'.format(
                    old_owner_as.user.username,
                    new_owner.username,
                    project.title,
                )
            },
            status=200,
        )


class ProjectInviteAPIMixin:
    """Validation helpers for project invite modification via API"""

    def validate(
        self, invite: Optional[ProjectInvite], request: HttpRequest, **kwargs
    ):
        if not invite:
            raise NotFound(
                'Invite not found (uuid={})'.format(kwargs['projectinvite'])
            )
        if (
            invite.role.name == PROJECT_ROLE_DELEGATE
            and not request.user.has_perm(
                'projectroles.update_project_delegate', invite.project
            )
        ):
            raise PermissionDenied(
                'User lacks permission to modify delegate invites'
            )
        if not invite.active:
            raise serializers.ValidationError('Invite is not active')


class ProjectInviteListAPIView(
    ProjectrolesAPIVersioningMixin, SODARAPIBaseProjectMixin, ListAPIView
):
    """
    List user invites for a project.

    Supports optional pagination by providing the ``page`` query string. This
    will return results in the Django Rest Framework ``PageNumberPagination``
    format.

    **URL:** ``/project/api/invites/list/{Project.sodar_uuid}``

    **Methods:** ``GET``

    **Parameters:**

    - ``page``: Page number for paginated results (int, optional)

    **Returns:** List or paginated dict of project invite details

    **Version Changes:**

    - ``2.0``: Replace ``issuer`` field user serializer with user UUID
    """

    pagination_class = SODARPageNumberPagination
    permission_required = 'projectroles.invite_users'
    serializer_class = ProjectInviteSerializer

    def get_queryset(self):
        return ProjectInvite.objects.filter(
            project=self.get_project(), active=True
        ).order_by('pk')


class ProjectInviteRetrieveAPIView(
    ProjectrolesAPIVersioningMixin, SODARAPIGenericProjectMixin, RetrieveAPIView
):
    """
    Retrieve a project invite.

    **URL:** ``/project/api/invites/retrieve/{ProjectInvite.sodar_uuid}``

    **Methods:** ``GET``

    **Returns:**

    - ``date_created``: Creation datetime string (YYYY-MM-DDThh:mm:ssZ)
    - ``date_expire``: Expiry datetime string (YYYY-MM-DDThh:mm:ssZ)
    - ``message``: Optional invite message (string)
    - ``issuer``: UUID of issuing user (string)
    - ``email``: Email of invited user (string)
    - ``project``: Project UUID (string)
    - ``role``: Role for invided user (string, e.g. "project contributor")
    - ``sodar_uuid``: Invite UUID (string)

    **Version Changes:**

    - ``2.0``: Add view
    """

    lookup_field = 'sodar_uuid'
    lookup_url_kwarg = 'projectinvite'
    permission_required = 'projectroles.invite_users'
    serializer_class = ProjectInviteSerializer

    def get(self, request, *args, **kwargs):
        """Override get() to check for API version"""
        if parse_version(request.version) < VERSION_2_0:
            raise NotAcceptable(VIEW_NOT_ACCEPTABLE_VERSION_MSG)
        return super().get(request, *args, **kwargs)


class ProjectInviteCreateAPIView(
    ProjectrolesAPIVersioningMixin, SODARAPIGenericProjectMixin, CreateAPIView
):
    """
    Create a project invite.

    **URL:** ``/project/api/invites/create/{Project.sodar_uuid}``

    **Methods:** ``POST``

    **Parameters:**

    - ``email``: User email (string)
    - ``role``: Desired role for user (string, e.g. "project contributor")
    """

    permission_required = 'projectroles.invite_users'
    serializer_class = ProjectInviteSerializer

    def perform_create(self, serializer):
        project = self.get_project()
        user_email = self.request.data.get('email')
        existing_inv = ProjectInvite.objects.filter(
            project=project,
            email=user_email,
            active=True,
            date_expire__gt=timezone.now(),
        ).first()
        if existing_inv:
            raise serializers.ValidationError(
                INVITE_EXISTS_MSG.format(
                    user_email=user_email, project_title=project.get_log_title()
                )
            )
        if project.parent:
            parent_inv = ProjectInvite.objects.filter(
                email=user_email,
                active=True,
                date_expire__gt=timezone.now(),
                project__in=project.get_parents(),
            ).first()
            if parent_inv:
                raise serializers.ValidationError(
                    INVITE_EXISTS_MSG.format(
                        user_email=user_email,
                        project_title=parent_inv.project.get_log_title(),
                    )
                )
        return super().perform_create(serializer)


class ProjectInviteRevokeAPIView(
    ProjectInviteMixin,
    ProjectInviteAPIMixin,
    ProjectrolesAPIVersioningMixin,
    SODARAPIBaseProjectMixin,
    APIView,
):
    """
    Revoke a project invite.

    **URL:** ``/project/api/invites/revoke/{ProjectInvite.sodar_uuid}``

    **Methods:** ``POST``
    """

    permission_required = 'projectroles.invite_users'
    serializer_class = ProjectInviteSerializer

    def post(self, request, *args, **kwargs):
        """Handle invite revoking in a POST request"""
        invite = ProjectInvite.objects.filter(
            sodar_uuid=kwargs['projectinvite']
        ).first()
        self.validate(invite, request, **kwargs)
        invite = self.revoke_invite(invite, invite.project, request)
        return Response(
            {
                'detail': f'Invite revoked from email {invite.email} in '
                f'project "{invite.project.title}"'
            },
            status=200,
        )


class ProjectInviteResendAPIView(
    ProjectInviteMixin,
    ProjectInviteAPIMixin,
    ProjectrolesAPIVersioningMixin,
    SODARAPIBaseProjectMixin,
    APIView,
):
    """
    Resend email for a project invite.

    **URL:** ``/project/api/invites/resend/{ProjectInvite.sodar_uuid}``

    **Methods:** ``POST``
    """

    permission_required = 'projectroles.invite_users'
    serializer_class = ProjectInviteSerializer

    def post(self, request, *args, **kwargs):
        """Handle invite resending in a POST request"""
        invite = ProjectInvite.objects.filter(
            sodar_uuid=kwargs['projectinvite']
        ).first()
        self.validate(invite, request, **kwargs)
        self.handle_invite(invite, request, resend=True, add_message=False)
        return Response(
            {
                'detail': f'Invite resent from email {invite.email} in project '
                f'"{invite.project.title}"'
            },
            status=200,
        )


class AppSettingMixin:
    """Helpers for app setting API views"""

    @classmethod
    def get_and_validate_def(
        cls, plugin_name: str, setting_name: str, allowed_scopes: list[str]
    ) -> PluginAppSettingDef:
        """
        Return settings definition or raise a validation error.

        :param plugin_name: Name of app plugin for the setting (string)
        :param setting_name: Setting name (string)
        :param allowed_scopes: Allowed scopes for the setting (list)
        :return
        """
        try:
            s_def = app_settings.get_definition(
                name=setting_name, plugin_name=plugin_name
            )
        except Exception as ex:
            raise serializers.ValidationError(ex)
        if s_def.scope not in allowed_scopes:
            raise serializers.ValidationError('Invalid setting scope')
        return s_def

    @classmethod
    def get_request_value(cls, request: HttpRequest) -> str:
        """
        Return setting value from request.

        :param request: HTTPRequest object
        :return: String or None
        :raise: ValidationError if value is not set
        """
        if 'value' not in request.data:
            raise serializers.ValidationError(NO_VALUE_MSG)
        return request.data['value']

    @classmethod
    def check_project_perms(
        cls,
        setting_def: PluginAppSettingDef,
        project: Project,
        request: HttpRequest,
        setting_user: Optional[SODARUser],
    ):
        """
        Check permissions for project settings.

        :param setting_def: PluginAppSettingDef object
        :param project: Project object
        :param request: HttpRequest object
        :param setting_user: User object for the setting user or None
        """
        if setting_def.scope == APP_SETTING_SCOPE_PROJECT:
            if request.method == 'GET':
                perm = 'projectroles.view_project_settings'
            else:
                perm = 'projectroles.update_project_settings'
            if not request.user.has_perm(perm, project):
                raise PermissionDenied(
                    'User lacks permission to access PROJECT scope app '
                    'settings in this project'
                )
        elif setting_def.scope == APP_SETTING_SCOPE_PROJECT_USER:
            if not setting_user:
                raise serializers.ValidationError(
                    'No user given for PROJECT_USER setting'
                )
            if request.user != setting_user and not request.user.is_superuser:
                raise PermissionDenied(
                    'User is not allowed to update settings for other users'
                )
            if (
                request.method == 'POST'
                and not request.user.is_superuser
                and app_settings.get(APP_NAME, 'site_read_only')
            ):
                raise PermissionDenied(
                    'Site in read-only mode, operation not allowed'
                )

    @classmethod
    def _get_setting_obj(
        cls,
        plugin_name: Optional[str],
        setting_name: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ) -> AppSetting:
        """
        Return the database object for a setting. Returns None if not available.

        :param plugin_name: Name of the app plugin (string or None)
        :param setting_name: Setting name (string)
        :param project: Project object or None
        :param user: User object or None
        :return: AppSetting object
        :raise: AppSetting.DoesNotExist if not found
        :raise: AppSetting.MultipleObjectsReturned if the arguments are invalid
        """
        q_kwargs = {
            'name': setting_name,
            'project': project,
            'user': user,
        }
        if plugin_name == APP_NAME:
            q_kwargs['app_plugin'] = None
        else:
            q_kwargs['app_plugin__name'] = plugin_name
        return AppSetting.objects.get(**q_kwargs)

    @classmethod
    def get_setting_for_api(
        cls,
        plugin_name: Optional[str],
        setting_name: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ) -> AppSetting:
        """
        Return the database object for a setting for API serving. Will create
        the object if not yet created.

        :param plugin_name: Name of the app plugin (string or None)
        :param setting_name: Setting name (string)
        :param project: Project object or None
        :param user: User object or None
        :return: AppSetting object
        """
        try:
            return cls._get_setting_obj(
                plugin_name, setting_name, project, user
            )
        except AppSetting.DoesNotExist:
            try:
                app_settings.set(
                    plugin_name=plugin_name,
                    setting_name=setting_name,
                    value=app_settings.get_default(
                        plugin_name, setting_name, project=project, user=user
                    ),
                    project=project,
                    user=user,
                )
                return cls._get_setting_obj(
                    plugin_name, setting_name, project, user
                )
            except Exception as ex:
                raise serializers.ValidationError(ex)


class ProjectSettingRetrieveAPIView(
    ProjectrolesAPIVersioningMixin,
    SODARAPIBaseProjectMixin,
    AppSettingMixin,
    RetrieveAPIView,
):
    """
    API view for retrieving an app setting with the PROJECT or PROJECT_USER
    scope.

    **URL:** ``project/api/settings/retrieve/{Project.sodar_uuid}``

    **Methods:** ``GET``

    **Parameters:**

    - ``plugin_name``: Name of app plugin for the setting, use "projectroles" for projectroles settings (string)
    - ``setting_name``: Setting name (string)
    - ``user``: User UUID for a PROJECT_USER setting (string or None, optional)

    **Returns:**

    - ``plugin_name``: Name of app plugin for the setting (string)
    - ``project``: Project UUID or None
    - ``user``: User UUID or None
    - ``name``: Setting name (string)
    - ``type``: Setting type (string)
    - ``value``: Setting value
    - ``user_modifiable``: Modifiable by user (boolean)

    **Version Changes:**

    - ``2.0``: Replace ``user`` field user serializer with user UUID
    """

    # NOTE: Update project settings perm is checked manually
    permission_required = 'projectroles.view_project'
    serializer_class = AppSettingSerializer

    def get_object(self):
        plugin_name = self.request.GET.get('plugin_name')
        setting_name = self.request.GET.get('setting_name')

        # Get and validate definition
        s_def = self.get_and_validate_def(
            plugin_name,
            setting_name,
            [APP_SETTING_SCOPE_PROJECT, APP_SETTING_SCOPE_PROJECT_USER],
        )

        # Get project and user, check perms
        project = self.get_project()
        setting_user = None
        if self.request.GET.get('user'):
            setting_user = User.objects.filter(
                sodar_uuid=self.request.GET['user']
            ).first()
        self.check_project_perms(s_def, project, self.request, setting_user)

        # Return new object with default setting if not set
        return self.get_setting_for_api(
            plugin_name, setting_name, project, setting_user
        )


class ProjectSettingSetAPIView(
    ProjectrolesAPIVersioningMixin,
    SODARAPIBaseProjectMixin,
    AppSettingMixin,
    ProjectModifyPluginViewMixin,
    APIView,
):
    """
    API view for setting the value of an app setting with the PROJECT or
    PROJECT_USER scope.

    **URL:** ``project/api/settings/set/{Project.sodar_uuid}``

    **Methods:** ``POST``

    **Parameters:**

    - ``plugin_name``: Name of app plugin for the setting, use "projectroles" for projectroles settings (string)
    - ``setting_name``: Setting name (string)
    - ``value``: Setting value (string, may contain JSON for JSON settings)
    - ``user``: User UUID for a PROJECT_USER setting (string or None, optional)
    """

    http_method_names = ['post']
    # NOTE: Update project settings perm is checked manually
    permission_required = 'projectroles.view_project'
    serializer_class = AppSettingSerializer

    @transaction.atomic
    def post(self, request, *args, **kwargs):
        timeline = plugin_api.get_backend_api('timeline_backend')
        plugin_name = request.data.get('plugin_name')
        setting_name = request.data.get('setting_name')

        # Get and validate definition
        s_def = self.get_and_validate_def(
            plugin_name,
            setting_name,
            [APP_SETTING_SCOPE_PROJECT, APP_SETTING_SCOPE_PROJECT_USER],
        )
        if (
            s_def.scope == APP_SETTING_SCOPE_PROJECT
            and not s_def.user_modifiable
        ):
            raise PermissionDenied(USER_MODIFIABLE_MSG)
        # Get value
        value = self.get_request_value(request)

        # Get project and user, check perms
        project = self.get_project()
        setting_user = None
        if request.data.get('user'):
            setting_user = User.objects.filter(
                sodar_uuid=request.data['user']
            ).first()
        self.check_project_perms(s_def, project, request, setting_user)

        # Custom projectroles validation
        if (
            plugin_name == APP_NAME
            and setting_name == 'ip_allow_list'
            and value
        ):
            ips = [s.strip() for s in value.split(',')]
            for ip in ips:
                try:
                    if '/' in ip:
                        ip_network(ip)
                    else:
                        ip_address(ip)
                except ValueError as ex:
                    raise serializers.ValidationError(ex)
        if (
            plugin_name == APP_NAME
            and setting_name == 'category_public_stats'
            and (project.parent or project.is_project())
        ):
            raise serializers.ValidationError(CAT_PUBLIC_STATS_API_MSG)

        # Set setting value with validation, return possible errors
        try:
            old_value = app_settings.get(
                plugin_name, setting_name, project=project, user=setting_user
            )
            app_settings.set(
                plugin_name=plugin_name,
                setting_name=setting_name,
                value=value,
                project=project,
                user=setting_user,
            )
            # Call for additional actions for project creation/update in plugins
            if s_def.scope == APP_SETTING_SCOPE_PROJECT and (
                settings,
                'PROJECTROLES_ENABLE_MODIFY_API',
                False,
            ):
                args = [
                    plugin_name,
                    setting_name,
                    value,
                    old_value,
                    project,
                    setting_user,
                ]
                self.call_project_modify_api(
                    'perform_project_setting_update',
                    'revert_project_setting_update',
                    args,
                )
        except Exception as ex:
            raise serializers.ValidationError(ex)

        if timeline:
            tl_desc = f'set value of {plugin_name}.settings.{setting_name}'
            if setting_user:
                tl_desc += ' for user {{{}}}'.format('user')
            setting_obj = self._get_setting_obj(
                plugin_name, setting_name, project, setting_user
            )
            tl_extra_data = {'value': setting_obj.get_value()}
            tl_event = timeline.add_event(
                project=project,
                app_name=APP_NAME,
                user=request.user,
                event_name='app_setting_set_api',
                description=tl_desc,
                classified=True,
                extra_data=tl_extra_data,
                status_type=timeline.TL_STATUS_OK,
            )
            if setting_user:
                tl_event.add_object(setting_user, 'user', setting_user.username)
        return Response(
            {
                'detail': 'Set value of {}.settings.{} '
                '(project={}; user={})'.format(
                    plugin_name,
                    setting_name,
                    project.sodar_uuid,
                    setting_user.sodar_uuid if setting_user else None,
                )
            },
            status=200,
        )


class UserSettingRetrieveAPIView(
    ProjectrolesAPIVersioningMixin, AppSettingMixin, RetrieveAPIView
):
    """
    API view for retrieving an app setting with the USER scope.

    **URL:** ``project/api/settings/retrieve/user``

    **Methods:** ``GET``

    **Parameters:**

    - ``plugin_name``: Name of app plugin for the setting, use "projectroles" for projectroles settings (string)
    - ``setting_name``: Setting name (string)

    **Returns:**

    - ``plugin_name``: Name of app plugin for the setting (string)
    - ``project``: None
    - ``user``: User UUID
    - ``name``: Setting name (string)
    - ``type``: Setting type (string)
    - ``value``: Setting value
    - ``user_modifiable``: Modifiable by user (boolean)

    **Version Changes:**

    - ``2.0``: Replace ``user`` field user serializer with user UUID
    """

    serializer_class = AppSettingSerializer

    def get_object(self):
        if not self.request.user.is_authenticated:
            raise PermissionDenied(ANON_ACCESS_MSG)
        plugin_name = self.request.GET.get('plugin_name')
        setting_name = self.request.GET.get('setting_name')
        # Get and validate definition
        self.get_and_validate_def(
            plugin_name, setting_name, [APP_SETTING_SCOPE_USER]
        )
        # Return new object with default setting if not set
        return self.get_setting_for_api(
            plugin_name, setting_name, user=self.request.user
        )


class UserSettingSetAPIView(
    ProjectrolesAPIVersioningMixin, AppSettingMixin, APIView
):
    """
    API view for setting the value of an app setting with the USER scope. Only
    allows the user to set the value of their own settings.

    **URL:** ``project/api/settings/set/user``

    **Methods:** ``POST``

    **Parameters:**

    - ``plugin_name``: Name of app plugin for the setting, use "projectroles" for projectroles settings (string)
    - ``setting_name``: Setting name (string)
    - ``value``: Setting value (string, may contain JSON for JSON settings)
    """

    http_method_names = ['post']
    serializer_class = AppSettingSerializer

    def post(self, request, *args, **kwargs):
        if not request.user.is_authenticated:
            raise PermissionDenied(ANON_ACCESS_MSG)
        plugin_name = request.data.get('plugin_name')
        setting_name = request.data.get('setting_name')
        s_def = self.get_and_validate_def(
            plugin_name, setting_name, [APP_SETTING_SCOPE_USER]
        )
        if not s_def.user_modifiable:
            raise PermissionDenied(USER_MODIFIABLE_MSG)
        value = self.get_request_value(request)

        try:
            app_settings.set(
                plugin_name=plugin_name,
                setting_name=setting_name,
                value=value,
                user=request.user,
            )
        except Exception as ex:
            raise serializers.ValidationError(ex)
        return Response(
            {'detail': f'Set value of {plugin_name}.settings.{setting_name}'},
            status=200,
        )


class UserListAPIView(ProjectrolesAPIVersioningMixin, ListAPIView):
    """
    Return a list of all users on the site. Excludes system users, unless called
    with superuser access.

    Supports optional pagination by providing the ``page`` query string. This
    will return results in the Django Rest Framework ``PageNumberPagination``
    format.

    If ``PROJECTROLES_API_USER_DETAIL_RESTRICT`` is set True on the server, this
    view is only accessible by users who have a contributor role or above in at
    least one category or project.

    **URL:** ``/project/api/users/list``

    **Methods:** ``GET``

    **Parameters:**

    - ``include_system_users``: Include system users if True (bool, optional)
    - ``page``: Page number for paginated results (int, optional)

    **Returns:** List or paginated dict of serializers users (see ``CurrentUserRetrieveAPIView``)

    **Version Changes:**

    - ``1.1``: Add ``include_system_users`` parameter
    """

    lookup_field = 'project__sodar_uuid'
    pagination_class = SODARPageNumberPagination
    permission_classes = [IsAuthenticated]
    serializer_class = SODARUserSerializer

    def get_queryset(self):
        """
        Override get_queryset() to return users according to requesting user
        access.
        """
        inc_system = self.request.GET.get('include_system_users')
        version = parse_version(self.request.version)
        if inc_system and version < VERSION_1_1:
            raise NotAcceptable(USER_LIST_INCLUDE_VERSION_MSG)
        qs = User.objects.all().order_by('pk')
        if self.request.user.is_superuser or inc_system:
            return qs
        return qs.exclude(groups__name=SODAR_CONSTANTS['SYSTEM_USER_GROUP'])

    def get(self, request, *args, **kwargs):
        """Override get() to restrict access if set on server"""
        role_kw = {
            'user': request.user,
            'role__rank__lte': ROLE_RANKING[PROJECT_ROLE_CONTRIBUTOR],
        }
        if (
            getattr(settings, 'PROJECTROLES_API_USER_DETAIL_RESTRICT', False)
            and not request.user.is_superuser
            and RoleAssignment.objects.filter(**role_kw).count() == 0
        ):
            raise PermissionDenied(USER_LIST_RESTRICT_MSG)
        return super().get(request, *args, **kwargs)


class UserRetrieveAPIView(ProjectrolesAPIVersioningMixin, RetrieveAPIView):
    """
    Return user details for user with the given UUID.

    If ``PROJECTROLES_API_USER_DETAIL_RESTRICT`` is set True on the server, this
    view is only accessible by users who have a finder role or above in at least
    one category or project.

    **URL:** ``/project/api/users/{SODARUser.sodar_uuid}``

    **Methods:** ``GET``

    **Returns:**

    - ``additional_emails``: Additional verified email addresses for user (list of strings)
    - ``auth_type``: User authentication type (string)
    - ``email``: Email address of the user (string)
    - ``is_superuser``: Superuser status (boolean)
    - ``name``: Full name of the user (string)
    - ``sodar_uuid``: User UUID (string)
    - ``username``: Username of the user (string)

    **Version Changes:**

    - ``1.1``: Add view
    """

    permission_classes = [IsAuthenticated]
    serializer_class = SODARUserSerializer

    def get_object(self):
        try:
            return User.objects.get(sodar_uuid=self.kwargs.get('user'))
        except User.DoesNotExist:
            raise NotFound()

    def get(self, request, *args, **kwargs):
        if parse_version(request.version) < VERSION_1_1:
            raise NotAcceptable(VIEW_NOT_ACCEPTABLE_VERSION_MSG)
        if (
            getattr(settings, 'PROJECTROLES_API_USER_DETAIL_RESTRICT', False)
            and not request.user.is_superuser
            and RoleAssignment.objects.filter(user=request.user).count() == 0
        ):
            raise PermissionDenied(USER_LIST_RESTRICT_MSG)
        return super().get(request, *args, **kwargs)


class CurrentUserRetrieveAPIView(
    ProjectrolesAPIVersioningMixin, RetrieveAPIView
):
    """
    Return user details for user performing the request.

    **URL:** ``/project/api/users/current``

    **Methods:** ``GET``

    **Returns:**

    User details, see ``UserRetrieveAPIView``.

    **Version Changes:**

    - ``1.1``: Add ``auth_type`` field
    """

    permission_classes = [IsAuthenticated]
    serializer_class = SODARUserSerializer

    def get_object(self):
        return self.request.user


# TODO: Update this for new API base classes
@extend_schema(
    responses={
        '200': inline_serializer(
            'RemoteProjectGetResponse',
            fields={
                'users': serializers.JSONField(),
                'projects': serializers.JSONField(),
                'peer_sites': serializers.JSONField(),
                'app_settings': serializers.JSONField(),
            },
        )
    }
)
class RemoteProjectGetAPIView(RemoteSyncAPIVersioningMixin, APIView):
    """API view for retrieving remote projects from a source site"""

    permission_classes = (AllowAny,)  # We check the secret in get()

    def get(self, request, *args, **kwargs):
        remote_api = RemoteProjectAPI()
        secret = kwargs['secret']
        try:
            target_site = RemoteSite.objects.get(
                mode=SITE_MODE_TARGET, secret=secret
            )
        except RemoteSite.DoesNotExist:
            return Response('Remote site not found, unauthorized', status=401)
        # Pass request version to get_source_data()
        accept_header = request.headers.get('Accept')
        if accept_header and 'version=' in accept_header:
            req_version = accept_header[accept_header.find('version=') + 8 :]
        else:
            req_version = SYNC_API_DEFAULT_VERSION
        sync_data = remote_api.get_source_data(target_site, req_version)
        # Update access date for target site remote projects
        target_site.projects.all().update(date_access=timezone.now())
        return Response(sync_data, status=200)



import datetime

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import render
from django.views.generic.edit import DeleteView, FormView
from django.views.generic.list import ListView
from django.urls import reverse
from knox.models import AuthToken

# Projectroles dependency
from projectroles.models import RoleAssignment
from projectroles.views import LoggedInPermissionMixin

from tokens.forms import UserTokenCreateForm


# Local constants
TOKEN_CREATE_MSG = 'Token created.'
TOKEN_DELETE_MSG = 'Token deleted.'
TOKEN_CREATE_RESTRICT_MSG = (
    'Token creation is restricted to users with project access'
)


class UserTokenListView(LoginRequiredMixin, LoggedInPermissionMixin, ListView):
    """View for listing and accessing user API tokens"""

    model = AuthToken
    permission_required = 'tokens.view_list'
    template_name = 'tokens/token_list.html'

    def get_queryset(self):
        """Only allow access to this user's query set"""
        return AuthToken.objects.filter(user=self.request.user).order_by('-pk')

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        user = self.request.user
        context['token_create_enable'] = True
        context['token_create_msg'] = ''
        if (
            not user.is_superuser
            and getattr(settings, 'TOKENS_CREATE_PROJECT_USER_RESTRICT', False)
            and RoleAssignment.objects.filter(user=user).count() == 0
        ):
            context['token_create_enable'] = False
            context['token_create_msg'] = TOKEN_CREATE_RESTRICT_MSG
        return context


class UserTokenCreateView(
    LoginRequiredMixin, LoggedInPermissionMixin, FormView
):
    """View for API token creation"""

    form_class = UserTokenCreateForm
    permission_required = 'tokens.create'
    template_name = 'tokens/token_create.html'

    def form_valid(self, form):
        ttl = datetime.timedelta(hours=form.clean().get('ttl')) or None
        context = self.get_context_data()
        _, context['token'] = AuthToken.objects.create(self.request.user, ttl)
        messages.success(self.request, TOKEN_CREATE_MSG)
        return render(self.request, 'tokens/token_create_success.html', context)


class UserTokenDeleteView(
    LoginRequiredMixin, LoggedInPermissionMixin, DeleteView
):
    """View for API token deletion"""

    model = AuthToken
    permission_required = 'tokens.delete'
    template_name = 'tokens/token_confirm_delete.html'

    def get_success_url(self):
        messages.success(self.request, TOKEN_DELETE_MSG)
        return reverse('tokens:list')

    def get_queryset(self):
        """Only allow access to this user's query set"""
        return AuthToken.objects.filter(user=self.request.user)



"""Ajax API view tests for the timeline app"""

from datetime import datetime

from django.urls import reverse
from django.utils.timezone import localtime

# Projectroles dependency
from projectroles.tests.test_models import ProjectMixin, RoleAssignmentMixin
from projectroles.tests.test_views import (
    ViewTestBase,
    PROJECT_TYPE_CATEGORY,
    PROJECT_TYPE_PROJECT,
)

from timeline.models import DEFAULT_MESSAGES
from timeline.views_ajax import EventExtraDataMixin
from timeline.templatetags.timeline_tags import get_status_style
from timeline.tests.test_models import TimelineEventMixin


# Local constants
APP_NAME_PR = 'projectroles'


class TimelineAjaxViewTestBase(
    ProjectMixin, RoleAssignmentMixin, TimelineEventMixin, ViewTestBase
):
    """Base class for timeline Ajax API view tests"""

    @classmethod
    def _format_ts(cls, timestamp: datetime) -> str:
        """Format timestamp as an expected value from the Ajax API view"""
        return localtime(timestamp).strftime('%Y-%m-%d %H:%M:%S')

    def setUp(self):
        super().setUp()
        self.category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.owner_as_cat = self.make_assignment(
            self.category, self.user, self.role_owner
        )
        self.project = self.make_project(
            'TestProject', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )


class TestProjectEventDetailAjaxView(TimelineAjaxViewTestBase):
    """Tests for ProjectEventDetailAjaxView"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            self.project, APP_NAME_PR, self.user, 'project_create'
        )
        self.event_status_init = self.event.set_status(
            'INIT', DEFAULT_MESSAGES['INIT']
        )
        self.event_status_ok = self.event.set_status(
            'OK', DEFAULT_MESSAGES['OK'], extra_data={'test': 'test'}
        )
        self.url = reverse(
            'timeline:ajax_detail_project',
            kwargs={'timelineevent': self.event.sodar_uuid},
        )

    def test_get(self):
        """Test ProjectEventDetailAjaxView GET"""
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        expected = {
            'app': self.event.app,
            'name': self.event.event_name,
            'user': self.user.username,
            'timestamp': self._format_ts(self.event.get_timestamp()),
            'status': [
                {
                    'type': 'OK',
                    'class': get_status_style(self.event_status_ok),
                    'description': DEFAULT_MESSAGES['OK'],
                    'timestamp': self._format_ts(
                        self.event_status_ok.timestamp
                    ),
                    'extra_status_link': reverse(
                        'timeline:ajax_extra_status',
                        kwargs={
                            'eventstatus': self.event_status_ok.sodar_uuid,
                        },
                    ),
                },
                {
                    'type': 'INIT',
                    'class': get_status_style(self.event_status_init),
                    'description': DEFAULT_MESSAGES['INIT'],
                    'timestamp': self._format_ts(
                        self.event_status_init.timestamp
                    ),
                    'extra_status_link': None,
                },
            ],
        }
        self.assertEqual(response.data, expected)

    def test_get_no_user(self):
        """Test GET for event with no user"""
        self.event.user = None
        self.event.save()
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['user'], 'N/A')


class TestProjectEventExtraAjaxView(TimelineAjaxViewTestBase):
    """Tests for ProjectEventExtraAjaxView"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            self.project,
            APP_NAME_PR,
            self.user,
            'project_create',
            extra_data={'example_data': 'example_extra_data'},
        )
        self.event_status_init = self.event.set_status(
            'INIT', DEFAULT_MESSAGES['INIT']
        )
        self.event_status_ok = self.event.set_status(
            'OK', DEFAULT_MESSAGES['OK']
        )
        self.url = reverse(
            'timeline:ajax_extra_project',
            kwargs={'timelineevent': self.event.sodar_uuid},
        )

    def test_get(self):
        """Test ProjectEventExtraAjaxView GET"""
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        expected = {
            'app': self.event.app,
            'user': self.user.username,
            'extra': self.event.extra_data,
        }
        self.assertIn(expected['app'], str(response.data))
        self.assertIn(expected['user'], str(response.data))
        self.assertIn(expected['extra']['example_data'], str(response.data))

    def test_get_no_user(self):
        """Test GET for event with no user"""
        self.event.user = None
        self.event.save()
        with self.login(self.user):
            response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['user'], 'N/A')


class TestSiteEventDetailAjaxView(TimelineAjaxViewTestBase):
    """Tests for SiteEventDetailAjaxView"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(None, APP_NAME_PR, self.user, 'test_event')
        self.event_status_init = self.event.set_status(
            'INIT', DEFAULT_MESSAGES['INIT']
        )
        self.event_status_ok = self.event.set_status(
            'OK', DEFAULT_MESSAGES['OK'], extra_data={'test': 'test'}
        )

    def test_get(self):
        """Test SiteEventDetailAjaxView GET"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'timeline:ajax_detail_site',
                    kwargs={'timelineevent': self.event.sodar_uuid},
                ),
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'app': self.event.app,
            'name': self.event.event_name,
            'user': self.user.username,
            'timestamp': self._format_ts(self.event.get_timestamp()),
            'status': [
                {
                    'type': 'OK',
                    'class': get_status_style(self.event_status_ok),
                    'description': DEFAULT_MESSAGES['OK'],
                    'timestamp': self._format_ts(
                        self.event_status_ok.timestamp
                    ),
                    'extra_status_link': reverse(
                        'timeline:ajax_extra_status',
                        kwargs={
                            'eventstatus': self.event_status_ok.sodar_uuid,
                        },
                    ),
                },
                {
                    'type': 'INIT',
                    'class': get_status_style(self.event_status_init),
                    'description': DEFAULT_MESSAGES['INIT'],
                    'timestamp': self._format_ts(
                        self.event_status_init.timestamp
                    ),
                    'extra_status_link': None,
                },
            ],
        }
        self.assertEqual(response.data, expected)


class TestSiteEventExtraAjaxView(TimelineAjaxViewTestBase):
    """Tests for SiteEventExtraAjaxView"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            None,
            APP_NAME_PR,
            self.user,
            'test_event',
            extra_data={'example_data': 'example_extra_data'},
        )
        self.event_status_init = self.event.set_status(
            'INIT', DEFAULT_MESSAGES['INIT']
        )
        self.event_status_ok = self.event.set_status(
            'OK', DEFAULT_MESSAGES['OK']
        )

    def test_get(self):
        """Test SiteEventExtraAjaxView GET"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'timeline:ajax_extra_site',
                    kwargs={'timelineevent': self.event.sodar_uuid},
                ),
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'app': self.event.app,
            'name': self.event.event_name,
            'user': self.user.username,
            'timestamp': self._format_ts(self.event.get_timestamp()),
            'extra': self.event.extra_data,
        }
        self.assertIn(expected['app'], str(response.data))
        self.assertIn(expected['user'], str(response.data))


class TestEventStatusExtraAjaxView(
    TimelineAjaxViewTestBase, EventExtraDataMixin
):
    """Tests for EventStatusExtraAjaxView"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            self.project, APP_NAME_PR, self.user, 'test_event'
        )
        self.event_status_init = self.event.set_status(
            'INIT',
            DEFAULT_MESSAGES['INIT'],
            extra_data={'test': 'test'},
        )
        self.event_site = self.make_event(
            None, APP_NAME_PR, self.user, 'test_event_site'
        )
        self.event_site_status_init = self.event_site.set_status(
            'INIT',
            DEFAULT_MESSAGES['INIT'],
            extra_data={'test': 'test'},
        )

    def test_get_project(self):
        """Test EventStatusExtraAjaxView GET with project event"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'timeline:ajax_extra_status',
                    kwargs={
                        'eventstatus': self.event_status_init.sodar_uuid,
                    },
                ),
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'app': self.event.app,
            'name': self.event.event_name,
            'user': self.user.username,
            'timestamp': self._format_ts(self.event.get_timestamp()),
            'extra': self.get_event_extra(self.event, status=0)['extra'],
        }
        self.assertEqual(response.data, expected)

    def test_get_site(self):
        """Test GET with site event"""
        with self.login(self.user):
            response = self.client.get(
                reverse(
                    'timeline:ajax_extra_status',
                    kwargs={
                        'eventstatus': self.event_site_status_init.sodar_uuid,
                    },
                ),
            )
        self.assertEqual(response.status_code, 200)
        expected = {
            'app': self.event_site.app,
            'name': self.event_site.event_name,
            'user': self.user.username,
            'timestamp': self._format_ts(self.event_site.get_timestamp()),
            'extra': self.get_event_extra(self.event_site, status=0)['extra'],
        }
        self.assertEqual(response.data, expected)



"""UI views for the appalerts app"""

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse
from django.views.generic import ListView, View
from django.conf import settings

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.views import (
    LoginRequiredMixin,
    LoggedInPermissionMixin,
)

from appalerts.models import AppAlert


app_settings = AppSettingAPI()


# Local constants
APP_NAME_PR = 'projectroles'


class AppAlertListView(LoginRequiredMixin, LoggedInPermissionMixin, ListView):
    """App alert list view"""

    permission_required = 'appalerts.view_alerts'
    template_name = 'appalerts/alert_list.html'

    def get_queryset(self):
        status = self.kwargs.get('status', 'active')
        if status == 'dismissed':
            return AppAlert.objects.filter(
                user=self.request.user, active=False
            ).order_by('-pk')
        return AppAlert.objects.filter(
            user=self.request.user, active=True
        ).order_by('-pk')

    def get_paginate_by(self, queryset):
        status = self.kwargs.get('status', 'active')
        if status == 'dismissed':
            return getattr(settings, 'DEFAULT_PAGINATION', 15)
        return None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['dismissed'] = (
            True if self.kwargs.get('status') == 'dismissed' else False
        )
        context['read_only_disable'] = False
        if not self.request.user.is_superuser and app_settings.get(
            APP_NAME_PR, 'site_read_only'
        ):
            context['read_only_disable'] = True
        return context


class AppAlertLinkRedirectView(
    LoginRequiredMixin, LoggedInPermissionMixin, View
):
    """View for redirecting to alert link and dismissing alert"""

    permission_required = 'appalerts.view_alerts'

    def _handle_error(self, msg: str) -> HttpResponseRedirect:
        messages.error(self.request, msg)
        return redirect(reverse('appalerts:list'))

    def get(self, request, **kwargs):
        alert = AppAlert.objects.filter(
            sodar_uuid=kwargs.get('appalert')
        ).first()
        # Handle errors
        if not alert:
            return self._handle_error('Alert not found.')
        if request.user != alert.user:
            return self._handle_error('Alert assigned to different user.')
        if not alert.url:
            return self._handle_error('No URL found for alert.')
        # All OK = dismiss and redirect
        alert.active = False
        alert.save()
        return redirect(alert.url)



# Generated by Django 4.2.22 on 2025-06-10 12:18

from django.db import migrations


ROLE_NAME = 'project viewer'
ROLE_RANK = 45
ROLE_DESC = (
    'The project viewer role is a limited guest role. They can view general '
    'project details and its members, but they can not access all apps, data '
    'or files in the project.'
)


def populate_viewer_role(apps, schema_editor):
    """Populate Role object for viewer role"""
    Role = apps.get_model('projectroles', 'Role')
    Role.objects.create(name=ROLE_NAME, rank=ROLE_RANK, description=ROLE_DESC)


def revert_viewer_role(apps, schema_editor):
    """Revert Role object creation for viewer role"""
    Role = apps.get_model('projectroles', 'Role')
    viewer_role = Role.objects.filter(name=ROLE_NAME).first()
    if viewer_role:
        viewer_role.delete()


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0034_fix_project_star'),
    ]
    operations = [
        migrations.RunPython(
            code=populate_viewer_role,
            reverse_code=revert_viewer_role,
        ),
    ]



from django import forms
from django.contrib import admin

from db_file_storage.form_widgets import DBAdminClearableFileInput
from .models import File, Folder, HyperLink


class FileForm(forms.ModelForm):
    class Meta:
        model = File
        exclude = []
        widgets = {'file': DBAdminClearableFileInput}


class FileAdmin(admin.ModelAdmin):
    form = FileForm


admin.site.register(File)
admin.site.register(Folder)
admin.site.register(HyperLink)



"""Forms for the adminalerts app"""

from django import forms
from django.utils import timezone

# Projectroles dependency
from projectroles.forms import SODARModelForm, SODARPagedownWidget

from adminalerts.models import AdminAlert


# Local constants
EMAIL_HELP_CREATE = 'Send alert as email to all users on this site'
EMAIL_HELP_UPDATE = 'Send updated alert as email to all users on this site'


class AdminAlertForm(SODARModelForm):
    """Form for AdminAlert creation/updating"""

    send_email = forms.BooleanField(
        initial=True,
        label='Send alert as email',
        required=False,
    )

    class Meta:
        model = AdminAlert
        fields = [
            'message',
            'date_expire',
            'active',
            'require_auth',
            'send_email',
            'description',
        ]

    def __init__(self, current_user=None, *args, **kwargs):
        """Override for form initialization"""
        super().__init__(*args, **kwargs)
        self.current_user = current_user

        # Set date_expire properties
        # NOTE: "format" works in source but not in widget, any way to fix?
        self.fields['date_expire'].label = 'Expiry date'
        self.fields['date_expire'].widget = forms.widgets.DateInput(
            attrs={'type': 'date'}, format='%Y-%m-%d'
        )
        # Set description widget with preview
        self.fields['description'].widget = SODARPagedownWidget(
            attrs={'show_preview': True}
        )

        # Creation
        if not self.instance.pk:
            self.initial['date_expire'] = timezone.now() + timezone.timedelta(
                days=1
            )
            self.fields['send_email'].help_text = EMAIL_HELP_CREATE
        # Updating
        else:  # self.instance.pk
            # Set description value as raw markdown
            self.initial['description'] = self.instance.description.raw
            self.fields['send_email'].help_text = EMAIL_HELP_UPDATE
            # Sending email for update should be false by default
            self.initial['send_email'] = False

    def clean(self):
        """Custom form validation and cleanup"""
        # Don't allow alerts to expire in the past :)
        if self.cleaned_data.get('date_expire') <= timezone.now():
            self.add_error(
                'date_expire', 'Expiry date must be set in the future'
            )
        return self.cleaned_data

    def save(self, *args, **kwargs):
        """Override of form saving function"""
        obj = super().save(commit=False)
        obj.user = self.current_user
        obj.save()
        return obj



"""Tests for tasks in the projectroles app"""

from test_plus.test import TestCase

from projectroles.app_settings import AppSettingAPI
from projectroles.models import (
    Role,
    SODAR_CONSTANTS,
    ROLE_RANKING,
    RemoteSite,
    RemoteProject,
)
from projectroles.tasks import sync_remote_site_task, READ_ONLY_SKIP_MSG
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RemoteTargetMixin,
    RoleAssignmentMixin,
)


app_settings = AppSettingAPI()


# SODAR constants
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']

# Local constants
APP_NAME = 'projectroles'


class TestSyncRemoteSiteTask(
    ProjectMixin, RoleMixin, RemoteTargetMixin, RoleAssignmentMixin, TestCase
):
    """Tests for the sync_remote_site_task() task"""

    maxDiff = None

    def setUp(self):
        super().setUp()
        self.role_owner = Role.objects.get_or_create(
            name=PROJECT_ROLE_OWNER, rank=ROLE_RANKING[PROJECT_ROLE_OWNER]
        )[0]
        self.user = self.make_user('user')
        self.user.save()
        self.category = self.make_project(
            'top_category', PROJECT_TYPE_CATEGORY, None
        )
        self.make_assignment(self.category, self.user, self.role_owner)
        self.project_source = self.make_project(
            'test_project_source', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as_source = self.make_assignment(
            self.project_source, self.user, self.role_owner
        )
        self.project_target = self.make_project(
            'test_project_target', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as_target = self.make_assignment(
            self.project_target, self.user, self.role_owner
        )
        self.source_site, self.remote_projects = self.set_up_as_target(
            projects=[self.category, self.project_source]
        )

    def test_sync(self):
        """Test sync_remote_site_task()"""
        with self.assertLogs('projectroles.tasks', 'INFO') as cm:
            sync_remote_site_task()
            self.assertNotIn(READ_ONLY_SKIP_MSG, cm.output[0])  # No skip
        self.assertEqual(RemoteSite.objects.all().count(), 1)
        self.assertEqual(RemoteSite.objects.first().name, self.source_site.name)
        self.assertEqual(RemoteProject.objects.all().count(), 2)

    def test_sync_change_name(self):
        """Test sync_remote_site_task() after title change in source site"""
        sync_remote_site_task()
        self.project_source.title = 'new title'
        self.project_source.save()
        sync_remote_site_task()
        self.assertEqual(RemoteSite.objects.all().count(), 1)
        self.assertEqual(RemoteSite.objects.first().name, self.source_site.name)
        self.assertEqual(RemoteProject.objects.all().count(), 2)
        self.assertEqual(
            RemoteProject.objects.get(
                project=self.project_source
            ).project.title,
            'new title',
        )

    def test_sync_change_settings(self):
        """
        Test sync_remote_site_task() after app_settings changes in source site.
        """
        sync_remote_site_task()
        app_settings.set(
            APP_NAME, 'ip_restrict', True, project=self.project_source
        )
        sync_remote_site_task()
        self.assertEqual(RemoteSite.objects.all().count(), 1)
        self.assertEqual(RemoteSite.objects.first().name, self.source_site.name)
        self.assertEqual(RemoteProject.objects.all().count(), 2)
        self.assertEqual(
            app_settings.get(
                APP_NAME, 'ip_restrict', project=self.project_source
            ),
            True,
        )

    def test_sync_read_only(self):
        """Test sync_remote_site_task() with site read-only mode"""
        app_settings.set(APP_NAME, 'site_read_only', True)
        with self.assertLogs('projectroles.tasks', 'INFO') as cm:
            sync_remote_site_task()
            self.assertIn(READ_ONLY_SKIP_MSG, cm.output[0])



"""URL configuration for the adminalerts app"""

from django.urls import path

import adminalerts.views_ajax
from adminalerts import views

app_name = 'adminalerts'

urls_ui = [
    path(
        route='list',
        view=views.AdminAlertListView.as_view(),
        name='list',
    ),
    path(
        route='detail/<uuid:adminalert>',
        view=views.AdminAlertDetailView.as_view(),
        name='detail',
    ),
    path(
        route='create',
        view=views.AdminAlertCreateView.as_view(),
        name='create',
    ),
    path(
        route='update/<uuid:adminalert>',
        view=views.AdminAlertUpdateView.as_view(),
        name='update',
    ),
    path(
        route='delete/<uuid:adminalert>/delete',
        view=views.AdminAlertDeleteView.as_view(),
        name='delete',
    ),
]

urls_ajax = [
    path(
        route='ajax/active/toggle/<uuid:adminalert>',
        view=adminalerts.views_ajax.AdminAlertActiveToggleAjaxView.as_view(),
        name='ajax_active_toggle',
    ),
]

urlpatterns = urls_ui + urls_ajax



"""UI views for the timeline app"""

from django.conf import settings
from django.views.generic import ListView

# Projectroles dependency
from projectroles.models import Project
from projectroles.utils import get_display_name
from projectroles.views import (
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    ProjectContextMixin,
    ProjectPermissionMixin,
)

from timeline.models import TimelineEvent


# Local variables
DEFAULT_PAGINATION = 15


class EventTimelineMixin:
    """Mixin for common event timeline operations"""

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        project = None
        if self.kwargs.get('project'):
            project = Project.objects.filter(
                sodar_uuid=self.kwargs['project']
            ).first()
        if project:
            context['timeline_title'] = '{} Timeline'.format(
                get_display_name(project.type, title=True)
            )
            context['timeline_mode'] = 'project'
        else:
            context['timeline_title'] = 'Site-Wide Timeline Events'
            context['timeline_mode'] = 'site'
        return context

    def get_queryset(self):
        project_uuid = self.kwargs.get('project')
        set_kwargs = {'project__sodar_uuid': project_uuid}
        if (
            project_uuid
            and not self.request.user.has_perm(
                'timeline.view_classified_event', self.get_permission_object()
            )
        ) or (not project_uuid and not self.request.user.is_superuser):
            set_kwargs['classified'] = False
        return TimelineEvent.objects.filter(**set_kwargs).order_by('-pk')


class ProjectTimelineView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    ProjectContextMixin,
    ProjectPermissionMixin,
    EventTimelineMixin,
    ListView,
):
    """View for displaying timeline events for a project"""

    permission_required = 'timeline.view_timeline'
    template_name = 'timeline/timeline.html'
    model = TimelineEvent
    paginate_by = getattr(settings, 'TIMELINE_PAGINATION', DEFAULT_PAGINATION)


class SiteTimelineView(
    LoginRequiredMixin, LoggedInPermissionMixin, EventTimelineMixin, ListView
):
    """View for displaying timeline events for site-wide events"""

    permission_required = 'timeline.view_site_timeline'
    template_name = 'timeline/timeline_site.html'
    model = TimelineEvent
    paginate_by = getattr(settings, 'TIMELINE_PAGINATION', DEFAULT_PAGINATION)


class AdminTimelineView(LoginRequiredMixin, LoggedInPermissionMixin, ListView):
    """View for displaying timeline events for admin site view"""

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['timeline_title'] = 'All Timeline Events'
        context['timeline_mode'] = 'admin'
        return context

    def get_queryset(self):
        return TimelineEvent.objects.order_by('-pk')

    permission_required = 'timeline.view_site_admin'
    template_name = 'timeline/timeline_site.html'
    model = TimelineEvent
    paginate_by = getattr(settings, 'TIMELINE_PAGINATION', DEFAULT_PAGINATION)


class ObjectTimelineMixin:
    """Mixin for common object timeline operations"""

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['timeline_title'] = '{} Timeline'.format(
            self.kwargs['object_model']
        )
        context['timeline_mode'] = 'object'
        return context

    def get_queryset(self):
        project = None
        classified_perm = 'timeline.view_classified_site_event'
        if self.kwargs.get('project'):
            project = Project.objects.filter(
                sodar_uuid=self.kwargs['project']
            ).first()
            classified_perm = 'timeline.view_classified_event'
        queryset = TimelineEvent.objects.get_object_events(
            project=project,
            object_model=self.kwargs['object_model'],
            object_uuid=self.kwargs['object_uuid'],
        )
        if not self.request.user.has_perm(
            classified_perm, self.get_permission_object()
        ):
            queryset = queryset.filter(classified=False)
        return queryset


class ProjectObjectTimelineView(ObjectTimelineMixin, ProjectTimelineView):
    """View for displaying object timeline for a project"""


class SiteObjectTimelineView(ObjectTimelineMixin, SiteTimelineView):
    """View for displaying object timeline for site-wide events"""



"""UI tests for the tokens app"""

from datetime import timedelta

from django.test import override_settings
from django.urls import reverse
from django.utils import timezone

from knox.models import AuthToken

from selenium.webdriver.common.by import By

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_ui import UITestBase


app_settings = AppSettingAPI()


# SODAR constants
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']

# Local constants
APP_NAME_PR = 'projectroles'


class TestUserTokenListView(UITestBase):
    """Tests for UserTokenListView"""

    def setUp(self):
        super().setUp()
        # Create users
        self.superuser = self.make_user('superuser', True)
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user', False)

        # Create tokens
        self.token = AuthToken.objects.create(
            self.regular_user, timedelta(days=5)
        )[0]
        self.token_no_expiry = AuthToken.objects.create(
            self.regular_user, None
        )[0]
        self.url = reverse('tokens:list')

    def test_create_button(self):
        """Test visibility of create button"""
        self.assertFalse(app_settings.get(APP_NAME_PR, 'site_read_only'))
        self.assert_element_exists(
            [self.superuser, self.regular_user],
            self.url,
            'sodar-tk-create-btn',
            True,
        )

    def test_create_button_read_only(self):
        """Test visibility of create button with site read-only mode"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        self.assert_element_exists(
            [self.superuser], self.url, 'sodar-tk-create-btn', True
        )
        self.assert_element_exists(
            [self.regular_user], self.url, 'sodar-tk-create-btn', False
        )

    @override_settings(TOKENS_CREATE_PROJECT_USER_RESTRICT=True)
    def test_create_button_restrict(self):
        """Test create button with restriction"""
        self.login_and_redirect(self.regular_user, self.url)
        elem = self.selenium.find_element(By.ID, 'sodar-tk-create-btn')
        self.assertEqual(elem.get_attribute('disabled'), 'true')

    @override_settings(TOKENS_CREATE_PROJECT_USER_RESTRICT=True)
    def test_create_button_restrict_role(self):
        """Test create button with restriction as user with role"""
        category = self.make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self.make_assignment(category, self.regular_user, self.role_guest)
        self.login_and_redirect(self.regular_user, self.url)
        elem = self.selenium.find_element(By.ID, 'sodar-tk-create-btn')
        self.assertIsNone(elem.get_attribute('disabled'))

    @override_settings(TOKENS_CREATE_PROJECT_USER_RESTRICT=True)
    def test_create_button_restrict_superuser(self):
        """Test create button with restriction as superuser"""
        self.login_and_redirect(self.superuser, self.url)
        elem = self.selenium.find_element(By.ID, 'sodar-tk-create-btn')
        self.assertIsNone(elem.get_attribute('disabled'))

    def test_list_items(self):
        """Test visibility of items in token list"""
        expected = [(self.superuser, 0), (self.regular_user, 2)]
        self.assert_element_count(
            expected, self.url, 'sodar-tk-list-item', 'class'
        )
        self.assert_element_count(
            expected, self.url, 'sodar-tk-item-dropdown', 'class'
        )

    def test_list_items_read_only(self):
        """Test visibility of items in token list with site read-only mode"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        expected = [(self.superuser, 0), (self.regular_user, 2)]
        self.assert_element_count(
            expected, self.url, 'sodar-tk-list-item', 'class'
        )
        expected = [(self.superuser, 0), (self.regular_user, 0)]
        self.assert_element_count(
            expected, self.url, 'sodar-tk-item-dropdown', 'class'
        )

    def test_list_expiry(self):
        """Test token expiry dates in token list"""
        self.login_and_redirect(self.regular_user, self.url)
        items = self.selenium.find_elements(By.CLASS_NAME, 'sodar-tk-list-item')
        self.assertEqual(len(items), 2)
        expiry_time = timezone.localtime(self.token.expiry).strftime(
            '%Y-%m-%d %H:%M'
        )
        values = []
        for item in items:
            values.append(item.find_elements(By.TAG_NAME, 'td')[2].text)
        self.assertCountEqual(values, ['Never', expiry_time])



"""Django Rest Framework serializers for the timeline app"""

from rest_framework import serializers

# Projectroles dependency
from projectroles.serializers import (
    SODARModelSerializer,
    SODARProjectModelSerializer,
)

from timeline.models import (
    TimelineEvent,
    TimelineEventStatus,
    TimelineEventObjectRef,
)


class ExtraDataRepresentationMixin:
    """Mixin for controlling extra data visibility to users"""

    def to_representation(self, instance):
        ret = super().to_representation(instance)
        user = self.context['request'].user
        project = instance.get_project()
        if (
            project
            and not user.has_perm('timeline.view_event_extra_data', project)
            and not user.is_superuser
        ):
            ret.pop('extra_data')
        elif not project and not user.is_superuser:
            ret.pop('extra_data')
        return ret


class TimelineEventObjectRefSerializer(
    ExtraDataRepresentationMixin, serializers.ModelSerializer
):
    """
    Serializer for the TimelineEventObjectRef model. All fields are set to
    read-only.
    """

    event = serializers.SlugRelatedField(
        slug_field='sodar_uuid', read_only=True
    )

    class Meta:
        model = TimelineEventObjectRef
        fields = [
            'event',
            'label',
            'name',
            'object_model',
            'object_uuid',
            'extra_data',
            'sodar_uuid',
        ]
        read_only_fields = fields


class TimelineEventStatusSerializer(
    ExtraDataRepresentationMixin, SODARModelSerializer
):
    """
    Serializer for the TimelineEventStatus model. All fields are set to
    read-only.
    """

    event = serializers.SlugRelatedField(
        slug_field='sodar_uuid', read_only=True
    )

    class Meta:
        model = TimelineEventStatus
        fields = [
            'event',
            'timestamp',
            'status_type',
            'description',
            'extra_data',
            'sodar_uuid',
        ]
        read_only_fields = fields


class TimelineEventSerializer(
    ExtraDataRepresentationMixin, SODARProjectModelSerializer
):
    """
    Serializer for the TimelineEvent model. All fields are set to read-only.
    """

    user = serializers.SlugRelatedField(
        allow_null=True, slug_field='sodar_uuid', read_only=True
    )
    status_changes = TimelineEventStatusSerializer(many=True, read_only=True)
    event_objects = TimelineEventObjectRefSerializer(many=True, read_only=True)

    class Meta:
        model = TimelineEvent
        fields = [
            'project',
            'app',
            'plugin',
            'user',
            'event_name',
            'description',
            'extra_data',
            'classified',
            'status_changes',
            'event_objects',
            'sodar_uuid',
        ]
        read_only_fields = fields



"""REST API views for the samplesheets app"""

from rest_framework.generics import (
    ListCreateAPIView,
    RetrieveUpdateDestroyAPIView,
    GenericAPIView,
)
from rest_framework.renderers import JSONRenderer
from rest_framework.versioning import AcceptHeaderVersioning

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import PluginAPI
from projectroles.views_api import (
    SODARAPIGenericProjectMixin,
    SODARPageNumberPagination,
)

from filesfolders.models import Folder
from filesfolders.serializers import (
    FolderSerializer,
    FileSerializer,
    HyperLinkSerializer,
)
from filesfolders.views import (
    FilesfoldersTimelineMixin,
    FileServeMixin,
    TL_OBJ_TYPES,
    APP_NAME,
)


plugin_api = PluginAPI()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
FILESFOLDERS_API_MEDIA_TYPE = (
    'application/vnd.bihealth.sodar-core.filesfolders+json'
)
FILESFOLDERS_API_DEFAULT_VERSION = '2.0'
FILESFOLDERS_API_ALLOWED_VERSIONS = ['2.0']


# Base Classes and Mixins ------------------------------------------------------


class FilesfoldersAPIVersioningMixin:
    """
    Filesfolders API view versioning mixin for overriding media type and
    accepted versions.
    """

    class FilesfoldersAPIRenderer(JSONRenderer):
        media_type = FILESFOLDERS_API_MEDIA_TYPE

    class FilesfoldersAPIVersioning(AcceptHeaderVersioning):
        allowed_versions = FILESFOLDERS_API_ALLOWED_VERSIONS
        default_version = FILESFOLDERS_API_DEFAULT_VERSION

    renderer_classes = [FilesfoldersAPIRenderer]
    versioning_class = FilesfoldersAPIVersioning


class ListCreateAPITimelineMixin(FilesfoldersTimelineMixin):
    """
    Mixin that ties ListCreateAPIView:s to the SODAR timeline for filesfolders.
    """

    def perform_create(self, serializer):
        # Actually perform the creation
        super().perform_create(serializer)
        # Register creation with timeline
        self.add_item_modify_event(
            obj=serializer.instance, request=self.request, view_action='create'
        )


class RetrieveUpdateDestroyAPITimelineMixin(FilesfoldersTimelineMixin):
    """
    Mixin that ties RetrieveUpdateDestroyAPIView:s to the SODAR timeline for
    filesfolders.
    """

    def perform_update(self, serializer):
        # Collect update_attrs and old_data
        update_attrs = ['name', 'folder', 'description', 'flag']
        old_item = serializer.instance
        old_data = {}

        if old_item.__class__.__name__ == 'HyperLink':
            update_attrs.append('url')
        elif old_item.__class__.__name__ == 'File':
            update_attrs.append('public_url')

        for a in update_attrs:
            old_data[a] = getattr(old_item, a)

        # Actually perform the update
        super().perform_update(serializer)

        # Register update with timeline
        self.add_item_modify_event(
            obj=serializer.instance,
            request=self.request,
            view_action='update',
            update_attrs=update_attrs,
            old_data=old_data,
        )

    def perform_destroy(self, instance):
        instance.delete()
        timeline = plugin_api.get_backend_api('timeline_backend')

        # Add event in Timeline
        if timeline:
            obj_type = TL_OBJ_TYPES[instance.__class__.__name__]
            # Add event in Timeline
            tl_event = timeline.add_event(
                project=instance.project,
                app_name=APP_NAME,
                user=self.request.user,
                event_name=f'{obj_type}_delete',
                description=f'delete {obj_type} {{{obj_type}}}',
                status_type=timeline.TL_STATUS_OK,
            )
            tl_event.add_object(
                obj=instance,
                label=obj_type,
                name=(
                    instance.get_path()
                    if isinstance(instance, Folder)
                    else instance.name
                ),
            )


class ListCreatePermissionMixin:
    """Mixing adding get_permission_required() for list-create API views."""

    def get_permission_required(self):
        if self.request.method == 'POST':
            return 'filesfolders.add_data'
        else:
            return 'filesfolders.view_data'


class RetrieveUpdateDestroyPermissionMixin:
    """
    Mixing adding get_permission_required() for retrieve-update-destroy API
    views.
    """

    def get_permission_required(self):
        if self.request.method == 'GET':
            return 'filesfolders.view_data'
        else:
            obj = self.get_object()
            if obj.owner == self.request.user:
                return 'filesfolders.update_data_own'
            else:
                return 'filesfolders.update_data_all'


# API Views --------------------------------------------------------------------


class FolderListCreateAPIView(
    ListCreateAPITimelineMixin,
    ListCreatePermissionMixin,
    FilesfoldersAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    ListCreateAPIView,
):
    """
    List folders or create a folder.

    Supports optional pagination for listing by providing the ``page`` query
    string. This will return results in the Django Rest Framework
    ``PageNumberPagination`` format.

    **URL:** ``/files/api/folder/list-create/{Project.sodar_uuid}``

    **Methods:** ``GET``, ``POST``

    **Parameters for GET:**

    - ``page``: Page number for paginated results (int, optional)

    **Parameters for POST:**

    - ``name``: Folder name (string)
    - ``folder``: Parent folder UUID (string)
    - ``owner``: User UUID of folder owner (string)
    - ``flag``: Folder flag (string, optional)
    - ``description``: Folder description (string, optional)
    """

    pagination_class = SODARPageNumberPagination
    project_type = PROJECT_TYPE_PROJECT
    serializer_class = FolderSerializer


class FolderRetrieveUpdateDestroyAPIView(
    RetrieveUpdateDestroyAPITimelineMixin,
    RetrieveUpdateDestroyPermissionMixin,
    FilesfoldersAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    RetrieveUpdateDestroyAPIView,
):
    """
    Retrieve, update or destroy a folder.

    **URL:** ``/files/api/folder/retrieve-update-destroy/{Folder.sodar_uuid}``

    **Methods:** ``GET``, ``PUT``, ``PATCH``, ``DELETE``

    **Parameters for PUT and PATCH:**

    - ``name``: Folder name (string)
    - ``folder``: Parent folder UUID (string)
    - ``owner``: User UUID of folder owner (string)
    - ``flag``: Folder flag (string, optional)
    - ``description``: Folder description (string, optional)
    """

    lookup_field = 'sodar_uuid'
    lookup_url_kwarg = 'folder'
    project_type = PROJECT_TYPE_PROJECT
    serializer_class = FolderSerializer


class FileListCreateAPIView(
    ListCreateAPITimelineMixin,
    ListCreatePermissionMixin,
    FilesfoldersAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    ListCreateAPIView,
):
    """
    List files or upload a file. For uploads, the request must be made in the
    ``multipart`` format.

    Supports optional pagination for listing by providing the ``page`` query
    string. This will return results in the Django Rest Framework
    ``PageNumberPagination`` format.

    **URL:** ``/files/api/file/list-create/{Project.sodar_uuid}``

    **Methods:** ``GET``, ``POST``

    **Parameters for GET:**

    - ``page``: Page number for paginated results (int, optional)

    **Parameters for POST:**

    - ``name``: Folder name (string)
    - ``folder``: Parent folder UUID (string)
    - ``owner``: User UUID of folder owner (string)
    - ``flag``: Folder flag (string, optional)
    - ``description``: Folder description (string, optional)
    - ``public_url``: Allow creation of a publicly viewable URL (bool)
    - ``file``: File to be uploaded
    """

    pagination_class = SODARPageNumberPagination
    project_type = PROJECT_TYPE_PROJECT
    serializer_class = FileSerializer


class FileRetrieveUpdateDestroyAPIView(
    RetrieveUpdateDestroyAPITimelineMixin,
    RetrieveUpdateDestroyPermissionMixin,
    FilesfoldersAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    RetrieveUpdateDestroyAPIView,
):
    """
    Retrieve, update or destroy a file.

    **URL:** ``/files/api/file/retrieve-update-destroy/{File.sodar_uuid}``

    **Methods:** ``GET``, ``PUT``, ``PATCH``, ``DELETE``

    **Parameters for PUT and PATCH:**

    - ``name``: Folder name (string)
    - ``folder``: Parent folder UUID (string)
    - ``owner``: User UUID of folder owner (string)
    - ``flag``: Folder flag (string, optional)
    - ``description``: Folder description (string, optional)
    - ``public_url``: Allow creation of a publicly viewable URL (bool)
    - ``file``: File to be uploaded
    """

    lookup_field = 'sodar_uuid'
    lookup_url_kwarg = 'file'
    project_type = PROJECT_TYPE_PROJECT
    serializer_class = FileSerializer


class FileServeAPIView(
    FilesfoldersAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    FileServeMixin,
    GenericAPIView,
):
    """
    Serve the file content.

    **URL:** ``/files/api/file/serve/{File.sodar_uuid}``

    **Methods:** ``GET``
    """

    lookup_field = 'sodar_uuid'
    lookup_url_kwarg = 'file'
    permission_required = 'filesfolders.view_data'
    schema = None


class HyperLinkListCreateAPIView(
    ListCreateAPITimelineMixin,
    ListCreatePermissionMixin,
    FilesfoldersAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    ListCreateAPIView,
):
    """
    List hyperlinks or create a hyperlink.

    Supports optional pagination for listing by providing the ``page`` query
    string. This will return results in the Django Rest Framework
    ``PageNumberPagination`` format.

    **URL:** ``/files/api/hyperlink/list-create/{Project.sodar_uuid}``

    **Methods:** ``GET``, ``POST``

    **Parameters for GET:**

    - ``page``: Page number for paginated results (int, optional)

    **Parameters for POST:**

    - ``name``: Folder name (string)
    - ``folder``: Parent folder UUID (string)
    - ``owner``: User UUID of folder owner (string)
    - ``flag``: Folder flag (string, optional)
    - ``description``: Folder description (string, optional)
    - ``url``: URL for the link (string)
    """

    pagination_class = SODARPageNumberPagination
    project_type = PROJECT_TYPE_PROJECT
    serializer_class = HyperLinkSerializer


class HyperLinkRetrieveUpdateDestroyAPIView(
    RetrieveUpdateDestroyAPITimelineMixin,
    RetrieveUpdateDestroyPermissionMixin,
    FilesfoldersAPIVersioningMixin,
    SODARAPIGenericProjectMixin,
    RetrieveUpdateDestroyAPIView,
):
    """
    Retrieve, update or destroy a hyperlink.

    **URL:** ``/files/api/hyperlink/retrieve-update-destroy/{HyperLink.sodar_uuid}``

    **Methods:** ``GET``, ``PUT``, ``PATCH``, ``DELETE``

    **Parameters for PUT and PATCH:**

    - ``name``: Folder name (string)
    - ``folder``: Parent folder UUID (string)
    - ``owner``: User UUID of folder owner (string)
    - ``flag``: Folder flag (string, optional)
    - ``description``: Folder description (string, optional)
    - ``url``: URL for the link (string)
    """

    lookup_field = 'sodar_uuid'
    lookup_url_kwarg = 'hyperlink'
    project_type = PROJECT_TYPE_PROJECT
    serializer_class = HyperLinkSerializer



from django.urls import path

from siteinfo import views

app_name = 'siteinfo'

urlpatterns = [
    path(
        route='info',
        view=views.SiteInfoView.as_view(),
        name='info',
    ),
]



"""Ajax API view permission tests for the projectroles app"""

from django.test import override_settings
from django.urls import reverse

from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_models import RemoteSiteMixin, RemoteProjectMixin
from projectroles.tests.test_permissions import (
    ProjectPermissionTestBase,
    SiteAppPermissionTestBase,
)

# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
REMOTE_LEVEL_READ_ROLES = SODAR_CONSTANTS['REMOTE_LEVEL_READ_ROLES']


class TestProjectListAjaxViews(ProjectPermissionTestBase):
    """Tests for project list Ajax view permissions"""

    def test_get_project_list(self):
        """Test ProjectListAjaxView GET"""
        url = reverse('projectroles:ajax_project_list')
        self.assert_response(url, self.auth_users, 200)
        self.assert_response(url, self.anonymous, 403)
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(url, self.user_no_roles, 200)
            self.assert_response(url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_project_list_anon(self):
        """Test ProjectListAjaxView GET with anonymous access"""
        url = reverse('projectroles:ajax_project_list')
        self.assert_response(url, self.anonymous, 200)
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(url, self.no_role_users, 200)

    def test_get_project_list_read_only(self):
        """Test ProjectListAjaxView GET with site read-only mode"""
        self.set_site_read_only()
        url = reverse('projectroles:ajax_project_list')
        self.assert_response(url, self.auth_users, 200)
        self.assert_response(url, self.anonymous, 403)
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(url, self.user_no_roles, 200)
            self.assert_response(url, self.anonymous, 403)

    def test_get_project_list_column(self):
        """Test ProjectListColumnAjaxView GET"""
        url = reverse('projectroles:ajax_project_list_columns')
        data = {'projects': [str(self.project.sodar_uuid)]}
        req_kwargs = {'content_type': 'application/json'}
        self.assert_response(
            url,
            self.auth_users,
            200,
            method='POST',
            data=data,
            req_kwargs=req_kwargs,
        )
        self.assert_response(
            url,
            self.anonymous,
            403,
            method='POST',
            data=data,
            req_kwargs=req_kwargs,
        )
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                url,
                self.user_no_roles,
                200,
                method='POST',
                data=data,
                req_kwargs=req_kwargs,
            )
            self.assert_response(
                url,
                self.anonymous,
                403,
                method='POST',
                data=data,
                req_kwargs=req_kwargs,
            )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_project_list_column_anon(self):
        """Test ProjectListColumnAjaxView GET with anonymous access"""
        url = reverse('projectroles:ajax_project_list_columns')
        data = {'projects': [str(self.project.sodar_uuid)]}
        req_kwargs = {'content_type': 'application/json'}
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                url,
                self.no_role_users,
                200,
                method='POST',
                data=data,
                req_kwargs=req_kwargs,
            )

    def test_get_project_list_column_read_only(self):
        """Test ProjectListColumnAjaxView GET with site read-only mode"""
        self.set_site_read_only()
        url = reverse('projectroles:ajax_project_list_columns')
        data = {'projects': [str(self.project.sodar_uuid)]}
        req_kwargs = {'content_type': 'application/json'}
        self.assert_response(
            url,
            self.auth_users,
            200,
            method='POST',
            data=data,
            req_kwargs=req_kwargs,
        )
        self.assert_response(
            url,
            self.anonymous,
            403,
            method='POST',
            data=data,
            req_kwargs=req_kwargs,
        )

    def test_get_project_list_role(self):
        """Test ProjectListRoleAjaxView GET"""
        url = reverse('projectroles:ajax_project_list_roles')
        data = {'projects': [str(self.project.sodar_uuid)]}
        req_kwargs = {'content_type': 'application/json'}
        self.assert_response(
            url,
            self.auth_users,
            200,
            method='POST',
            data=data,
            req_kwargs=req_kwargs,
        )
        self.assert_response(
            url,
            self.anonymous,
            403,
            method='POST',
            data=data,
            req_kwargs=req_kwargs,
        )
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                url,
                self.user_no_roles,
                200,
                method='POST',
                data=data,
                req_kwargs=req_kwargs,
            )
            self.assert_response(
                url,
                self.anonymous,
                403,
                method='POST',
                data=data,
                req_kwargs=req_kwargs,
            )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_project_list_role_anon(self):
        """Test ProjectListRoleAjaxView GET with anonymous access"""
        url = reverse('projectroles:ajax_project_list_roles')
        data = {'projects': [str(self.project.sodar_uuid)]}
        req_kwargs = {'content_type': 'application/json'}
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                url,
                self.no_role_users,
                200,
                method='POST',
                data=data,
                req_kwargs=req_kwargs,
            )

    def test_get_project_list_role_read_only(self):
        """Test ProjectListRoleAjaxView GET with site read-only mode"""
        self.set_site_read_only()
        url = reverse('projectroles:ajax_project_list_roles')
        data = {'projects': [str(self.project.sodar_uuid)]}
        req_kwargs = {'content_type': 'application/json'}
        self.assert_response(
            url,
            self.auth_users,
            200,
            method='POST',
            data=data,
            req_kwargs=req_kwargs,
        )
        self.assert_response(
            url,
            self.anonymous,
            403,
            method='POST',
            data=data,
            req_kwargs=req_kwargs,
        )
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                url,
                self.user_no_roles,
                200,
                method='POST',
                data=data,
                req_kwargs=req_kwargs,
            )
            self.assert_response(
                url,
                self.anonymous,
                403,
                method='POST',
                data=data,
                req_kwargs=req_kwargs,
            )


class TestCategoryStatisticsAjaxView(ProjectPermissionTestBase):
    """Tests for CategoryStatisticsAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'projectroles:ajax_stats_category',
            kwargs={'project': self.category.sodar_uuid},
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
        ]
        self.bad_users = [self.user_no_roles, self.anonymous]

    def test_get(self):
        """Test CategoryStatisticsAjaxView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(self.url, self.user_no_roles, 200)
            self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        for role in self.guest_roles:
            self.project.set_public_access(role)  # Public access in child
            self.assert_response(
                self.url, [self.user_no_roles, self.anonymous], 200
            )

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)

    def test_get_public_stats(self):
        """Test GET with category public stats"""
        self.set_category_public_stats(self.category)
        self.assert_response(self.url, self.auth_users, 200)
        self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_public_stats_anon(self):
        """Test GET with category public stats and anonymous access"""
        self.set_category_public_stats(self.category)
        self.assert_response(self.url, self.all_users, 200)


class TestProjectStarringAjaxView(ProjectPermissionTestBase):
    """Tests for ProjectStarringAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'projectroles:ajax_star',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.url_cat = reverse(
            'projectroles:ajax_star',
            kwargs={'project': self.category.sodar_uuid},
        )

    def test_post(self):
        """Test ProjectStarringAjaxView POST"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
        ]
        bad_users = [self.user_finder_cat, self.user_no_roles, self.anonymous]
        self.assert_response(self.url, good_users, 200, method='POST')
        self.assert_response(self.url, bad_users, 403, method='POST')
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                self.url, self.user_no_roles, 200, method='POST'
            )
            self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_post_anon(self):
        """Test POST with anonymous access"""
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                self.url, self.user_no_roles, 200, method='POST'
            )
            self.assert_response(self.url, self.anonymous, 401, method='POST')

    def test_post_block(self):
        """Test POST with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200, method='POST')
        self.assert_response(self.url, self.non_superusers, 403, method='POST')
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                self.url, self.non_superusers, 403, method='POST'
            )

    def test_post_read_only(self):
        """Test POST with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.superuser, 200, method='POST')
        self.assert_response(self.url, self.non_superusers, 403, method='POST')

    def test_post_category(self):
        """Test POST with category"""
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
        ]
        bad_users = [self.user_no_roles, self.anonymous]
        self.assert_response(self.url_cat, good_users, 200, method='POST')
        self.assert_response(self.url_cat, bad_users, 403, method='POST')
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                self.url_cat, self.user_no_roles, 200, method='POST'
            )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_post_category_anon(self):
        """Test POST with category and anonymous access"""
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(
                self.url, self.user_no_roles, 200, method='POST'
            )
            self.assert_response(
                self.url_cat, self.anonymous, 401, method='POST'
            )

    def test_post_category_read_only(self):
        """Test POST with category and site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url_cat, self.superuser, 200, method='POST')
        self.assert_response(
            self.url_cat, self.non_superusers, 403, method='POST'
        )

    def test_post_category_public_stats(self):
        """Test POST with category and public stats"""
        self.set_category_public_stats(self.category)
        self.assert_response(self.url_cat, self.auth_users, 200, method='POST')
        self.assert_response(self.url_cat, self.anonymous, 403, method='POST')

    def test_post_category_public_stats_anon(self):
        """Test POST with category and public stats and anonymous access"""
        self.set_category_public_stats(self.category)
        self.assert_response(self.url_cat, self.anonymous, 403, method='POST')


class TestHomeStarringAjaxView(SiteAppPermissionTestBase):
    """Tests for HomeStarringAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse('projectroles:ajax_star_home')
        self.post_data = {'value': '1'}

    def test_post(self):
        """Test HomeStarringAjaxView POST"""
        self.assert_response(
            self.url,
            [self.superuser, self.regular_user],
            200,
            method='POST',
            data=self.post_data,
        )
        self.assert_response(
            self.url, self.anonymous, 403, method='POST', data=self.post_data
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_post_anon(self):
        """Test POST with anonymous access"""
        self.assert_response(
            self.url, self.anonymous, 403, method='POST', data=self.post_data
        )


class TestRemoteProjectAccessAjaxView(
    RemoteSiteMixin, RemoteProjectMixin, ProjectPermissionTestBase
):
    """Tests for RemoteProjectAccessAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.remote_site = self.make_site(
            name='RemoteSite',
            url='https://remote.site',
            mode=SITE_MODE_TARGET,
        )
        self.remote_project = self.make_remote_project(
            project_uuid=self.project.sodar_uuid,
            site=self.remote_site,
            level=REMOTE_LEVEL_READ_ROLES,
            project=self.project,
        )
        self.url = (
            reverse(
                'projectroles:ajax_remote_access',
                kwargs={'project': self.project.sodar_uuid},
            )
            + f'?rp={self.remote_project.sodar_uuid}'
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
        ]
        self.bad_users = [
            self.user_finder_cat,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test RemoteProjectAccessAjaxView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(self.url, self.user_no_roles, 200)
            self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(self.url, self.no_role_users, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)


class TestSidebarContentAjaxView(ProjectPermissionTestBase):
    """Tests for SidebarContentAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'projectroles:ajax_sidebar',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.url_cat = reverse(
            'projectroles:ajax_sidebar',
            kwargs={'project': self.category.sodar_uuid},
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
        ]
        self.bad_users = [
            self.user_finder_cat,
            self.user_no_roles,
            self.anonymous,
        ]
        self.good_users_cat = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
        ]
        self.bad_users_cat = [self.user_no_roles, self.anonymous]

    def test_get(self):
        """Test SidebarContentAjaxView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(self.url, self.user_no_roles, 200)
            self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(self.url, self.no_role_users, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)

    def test_get_category(self):
        """Test GET with category"""
        self.assert_response(self.url_cat, self.good_users_cat, 200)
        self.assert_response(self.url_cat, self.bad_users_cat, 403)
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(self.url_cat, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_category_anon(self):
        """Test GET with category and anonymous access"""
        for role in self.guest_roles:
            self.project.set_public_access(role)
            self.assert_response(self.url_cat, self.no_role_users, 200)

    def test_get_category_read_only(self):
        """Test GET with category and site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url_cat, self.good_users_cat, 200)
        self.assert_response(self.url_cat, self.bad_users_cat, 403)


class TestSiteReadOnlySettingAjaxView(ProjectPermissionTestBase):
    """Tests for SiteReadOnlySettingAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse('projectroles:ajax_settings_site_read_only')

    def test_get(self):
        """Test SiteReadOnlySettingAjaxView GET"""
        self.assert_response(self.url, self.auth_users, 200)
        self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response(self.url, self.no_role_users, 200)


class TestUserDropdownContentAjaxView(ProjectPermissionTestBase):
    """Tests for UserDropdownContentAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse('projectroles:ajax_user_dropdown')

    def test_get(self):
        """Test UserDropdownContentAjaxView GET"""
        self.assert_response(self.url, self.auth_users, 200)
        self.assert_response(self.url, self.anonymous, 403)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.auth_users, 200)
        self.assert_response(self.url, self.anonymous, 403)


class TestUserAjaxViews(ProjectPermissionTestBase):
    """Tests for user Ajax view permissions"""

    def test_get_current_user(self):
        """Test CurrentUserRetrieveAjaxView GET"""
        url = reverse('projectroles:ajax_user_current')
        self.assert_response(url, self.auth_users, 200)
        self.assert_response(url, self.anonymous, 403)

    def test_get_current_user_read_only(self):
        """Test CurrentUserRetrieveAjaxView GET with site read-only mode"""
        self.set_site_read_only()
        url = reverse('projectroles:ajax_user_current')
        self.assert_response(url, self.auth_users, 200)
        self.assert_response(url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_LOCAL_USERS=True)
    def test_get_autocomplete_ajax(self):
        """Test UserAutocompleteAjaxView GET"""
        url = reverse('projectroles:ajax_autocomplete_user')
        self.assert_response(url, self.auth_users, 200)
        self.assert_response(url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_LOCAL_USERS=True)
    def test_get_autocomplete_ajax_read_only(self):
        """Test UserAutocompleteAjaxView GET with site read-only mode"""
        self.set_site_read_only()
        url = reverse('projectroles:ajax_autocomplete_user')
        self.assert_response(url, self.auth_users, 200)
        self.assert_response(url, self.anonymous, 403)



"""UI tests for the filesfolders app"""

from urllib.parse import urlencode

from django.urls import reverse

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.models import AppSetting, SODAR_CONSTANTS
from projectroles.tests.test_ui import UITestBase
from projectroles.utils import build_secret

from filesfolders.tests.test_models import (
    FolderMixin,
    FileMixin,
    HyperLinkMixin,
)


app_settings = AppSettingAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME = 'filesfolders'


class TestProjectFileView(FolderMixin, FileMixin, HyperLinkMixin, UITestBase):
    """Tests for ProjectFileView UI"""

    def setUp(self):
        super().setUp()
        app_settings.set(
            APP_NAME, 'allow_public_links', True, project=self.project
        )
        self.file_content = bytes('content'.encode('utf-8'))
        self.secret_file_owner = build_secret()
        self.secret_file_contributor = build_secret()
        # Folder created by project owner
        self.folder_owner = self.make_folder(
            name='folder_owner',
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='',
        )
        # Folder created by project contributor
        self.folder_contributor = self.make_folder(
            name='folder_contributor',
            project=self.project,
            folder=None,
            owner=self.user_contributor,
            description='',
        )
        # File uploaded by project owner
        self.file_owner = self.make_file(
            name='file_owner.txt',
            file_name='file_owner.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='',
            public_url=True,  # NOTE: Public URL OK
            secret=self.secret_file_owner,
        )
        # File uploaded by project contributor
        self.file_contributor = self.make_file(
            name='file_contributor.txt',
            file_name='file_contributor.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.user_contributor,
            description='',
            public_url=False,  # NOTE: No public URL
            secret=self.secret_file_contributor,
        )
        # HyperLink added by project owner
        self.hyperlink_owner = self.make_hyperlink(
            name='Owner link',
            url='https://www.bihealth.org/',
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='',
        )
        # HyperLink added by project contributor
        self.hyperlink_contrib = self.make_hyperlink(
            name='Contributor link',
            url='http://www.google.com/',
            project=self.project,
            folder=None,
            owner=self.user_contributor,
            description='',
        )
        self.url = reverse(
            'filesfolders:list', kwargs={'project': self.project.sodar_uuid}
        )

    def test_readme(self):
        """Test rendering readme if it has been uploaded to the folder"""
        # Init readme file
        self.readme_file = self.make_file(
            name='readme.txt',
            file_name='readme.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='',
            public_url=False,
            secret='xxxxxxxxx',
        )
        expected_true = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.assert_element_exists(
            expected_true, self.url, 'sodar-ff-readme-card', True
        )

    def test_ops_dropdown(self):
        """Test operations dropdown visibility"""
        expected_true = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        expected_false = [self.user_guest_cat, self.user_guest]
        self.assert_element_exists(
            expected_true, self.url, 'sodar-ff-file-ops-dropdown', True
        )
        self.assert_element_exists(
            expected_false, self.url, 'sodar-ff-file-ops-dropdown', False
        )

    def test_ops_dropdown_archive(self):
        """Test ops dropdown visibility for archived project"""
        self.project.set_archive()
        expected_true = [self.superuser]
        expected_false = [
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.assert_element_exists(
            expected_true, self.url, 'sodar-ff-file-ops-dropdown', True
        )
        self.assert_element_exists(
            expected_false, self.url, 'sodar-ff-file-ops-dropdown', False
        )

    def test_file_dropdown(self):
        """Test file dropdown visibility"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 0),  # Does not own files
            (self.user_guest_cat, 0),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 1),
            (self.user_guest, 0),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-ff-file-dropdown', 'class'
        )

    def test_file_dropdown_archive(self):
        """Test file dropdown visibility for archived project"""
        self.project.set_archive()
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 0),
            (self.user_delegate_cat, 0),
            (self.user_contributor_cat, 0),
            (self.user_guest_cat, 0),
            (self.user_owner, 0),
            (self.user_delegate, 0),
            (self.user_contributor, 0),
            (self.user_guest, 0),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-ff-file-dropdown', 'class'
        )

    def test_folder_dropdown(self):
        """Test folder dropdown visibility"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 0),  # Does not own folders
            (self.user_guest_cat, 0),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 1),
            (self.user_guest, 0),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-ff-folder-dropdown', 'class'
        )

    def test_folder_dropdown_archive(self):
        """Test folder dropdown visibility for archived project"""
        self.project.set_archive()
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 0),
            (self.user_delegate_cat, 0),
            (self.user_contributor_cat, 0),
            (self.user_guest_cat, 0),
            (self.user_owner, 0),
            (self.user_delegate, 0),
            (self.user_contributor, 0),
            (self.user_guest, 0),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-ff-folder-dropdown', 'class'
        )

    def test_hyperlink_dropdown(self):
        """Test hyperlink dropdown visibility"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 0),  # Does not own hyperlinks
            (self.user_guest_cat, 0),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 1),
            (self.user_guest, 0),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-ff-hyperlink-dropdown', 'class'
        )

    def test_hyperlink_dropdown_archive(self):
        """Test hyperlink dropdown visibility for archived project"""
        self.project.set_archive()
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 0),
            (self.user_delegate_cat, 0),
            (self.user_contributor_cat, 0),
            (self.user_guest_cat, 0),
            (self.user_owner, 0),
            (self.user_delegate, 0),
            (self.user_contributor, 0),
            (self.user_guest, 0),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-ff-hyperlink-dropdown', 'class'
        )

    def test_file_checkboxes(self):
        """Test batch file editing checkbox visibility"""
        expected = [
            (self.superuser, 6),
            (self.user_owner_cat, 6),
            (self.user_delegate_cat, 6),
            (self.user_contributor_cat, 0),
            (self.user_guest_cat, 0),
            (self.user_owner, 6),
            (self.user_delegate, 6),
            (self.user_contributor, 3),
            (self.user_guest, 0),
        ]
        self.assert_element_count(expected, self.url, 'sodar-ff-checkbox')

    def test_file_checkboxes_archive(self):
        """Test batch file editing checkbox visibility for archived project"""
        self.project.set_archive()
        expected = [
            (self.superuser, 6),
            (self.user_owner_cat, 0),
            (self.user_delegate_cat, 0),
            (self.user_contributor_cat, 0),
            (self.user_guest_cat, 0),
            (self.user_owner, 0),
            (self.user_delegate, 0),
            (self.user_contributor, 0),
            (self.user_guest, 0),
        ]
        self.assert_element_count(expected, self.url, 'sodar-ff-checkbox')

    def test_public_link(self):
        """Test public link visibility"""
        expected = [
            (self.superuser, 1),
            (self.user_owner_cat, 1),
            (self.user_delegate_cat, 1),
            (self.user_contributor_cat, 1),
            (self.user_guest_cat, 0),
            (self.user_owner, 1),
            (self.user_delegate, 1),
            (self.user_contributor, 1),
            (self.user_guest, 0),
        ]
        self.assert_element_count(expected, self.url, 'sodar-ff-link-public')

    def test_public_link_disable(self):
        """Test public link visibility if allow_public_links is set to False"""
        setting = AppSetting.objects.get(
            project=self.project.pk,
            app_plugin__name=APP_NAME,
            name='allow_public_links',
        )
        setting.value = 0
        setting.save()
        expected = [
            (self.superuser, 0),
            (self.user_owner_cat, 0),
            (self.user_delegate_cat, 0),
            (self.user_contributor_cat, 0),
            (self.user_guest_cat, 0),
            (self.user_owner, 0),
            (self.user_delegate, 0),
            (self.user_contributor, 0),
            (self.user_guest, 0),
        ]
        self.assert_element_count(expected, self.url, 'sodar-ff-link-public')

    def test_public_link_archive(self):
        """Test public link visibility for archived project"""
        self.project.set_archive()
        expected = [
            (self.superuser, 1),
            (self.user_owner_cat, 1),
            (self.user_delegate_cat, 1),
            (self.user_contributor_cat, 1),
            (self.user_guest_cat, 0),
            (self.user_owner, 1),
            (self.user_delegate, 1),
            (self.user_contributor, 1),
            (self.user_guest, 0),
        ]
        self.assert_element_count(expected, self.url, 'sodar-ff-link-public')

    def test_item_flags(self):
        """Test item flagging"""
        # Set up flags
        self.file_owner.flag = 'IMPORTANT'
        self.file_owner.save()
        self.folder_contributor.flag = 'FLAG'
        self.folder_contributor.save()
        self.hyperlink_contrib.flag = 'REVOKED'
        self.hyperlink_contrib.save()
        expected = [
            (self.superuser, 3),
            (self.user_owner_cat, 3),
            (self.user_delegate_cat, 3),
            (self.user_contributor_cat, 3),
            (self.user_guest_cat, 3),
            (self.user_owner, 3),
            (self.user_delegate, 3),
            (self.user_contributor, 3),
            (self.user_guest, 3),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-ff-flag-icon', 'class'
        )


class TestSearch(FolderMixin, FileMixin, HyperLinkMixin, UITestBase):
    """Tests for the project search UI functionalities"""

    def setUp(self):
        super().setUp()
        self.file_content = bytes('content'.encode('utf-8'))
        self.secret_file_owner = build_secret()
        self.secret_file_contributor = build_secret()
        # Folder created by project owner
        self.folder_owner = self.make_folder(
            name='folder_owner',
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='description',
        )
        # File created by project contributor
        self.folder_contributor = self.make_folder(
            name='folder_contributor',
            project=self.project,
            folder=None,
            owner=self.user_contributor,
            description='description',
        )
        # File uploaded by project owner
        self.file_owner = self.make_file(
            name='file_owner.txt',
            file_name='file_owner.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='description',
            public_url=True,  # NOTE: Public URL OK
            secret=self.secret_file_owner,
        )
        # File uploaded by project contributor
        self.file_contributor = self.make_file(
            name='file_contributor.txt',
            file_name='file_contributor.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.user_contributor,
            description='description',
            public_url=False,  # NOTE: No public URL
            secret=self.secret_file_contributor,
        )
        # HyperLink added by project owner
        self.hyperlink_owner = self.make_hyperlink(
            name='Owner link',
            url='https://www.bihealth.org/',
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='description',
        )
        # HyperLink added by project contributor
        self.hyperlink_contrib = self.make_hyperlink(
            name='Contributor link',
            url='http://www.google.com/',
            project=self.project,
            folder=None,
            owner=self.user_contributor,
            description='description',
        )
        self.url = reverse('projectroles:search')

    def test_search_results(self):
        """Test search items visibility according to user permissions"""
        expected = [
            (self.superuser, 6),
            (self.user_owner_cat, 6),
            (self.user_delegate_cat, 6),
            (self.user_contributor_cat, 6),
            (self.user_guest_cat, 6),
            (self.user_viewer_cat, 0),
            (self.user_finder_cat, 0),
            (self.user_owner, 6),
            (self.user_delegate, 6),
            (self.user_contributor, 6),
            (self.user_guest, 6),
            (self.user_viewer, 0),
            (self.user_no_roles, 0),
        ]
        url = self.url + '?' + urlencode({'s': 'description'})
        self.assert_element_count(expected, url, 'sodar-ff-search-item')

    def test_search_type_file(self):
        """Test search items visibility with 'file' type"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 2),
            (self.user_guest_cat, 2),
            (self.user_viewer_cat, 0),
            (self.user_finder_cat, 0),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 2),
            (self.user_guest, 2),
            (self.user_viewer, 0),
            (self.user_no_roles, 0),
        ]
        url = self.url + '?' + urlencode({'s': 'file type:file'})
        self.assert_element_count(expected, url, 'sodar-ff-search-item')

    def test_search_type_folder(self):
        """Test search items visibility with 'folder' type"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 2),
            (self.user_guest_cat, 2),
            (self.user_viewer_cat, 0),
            (self.user_finder_cat, 0),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 2),
            (self.user_guest, 2),
            (self.user_viewer, 0),
            (self.user_no_roles, 0),
        ]
        url = self.url + '?' + urlencode({'s': 'folder type:folder'})
        self.assert_element_count(expected, url, 'sodar-ff-search-item')

    def test_search_type_link(self):
        """Test search items visibility with 'link' as type"""
        expected = [
            (self.superuser, 2),
            (self.user_owner_cat, 2),
            (self.user_delegate_cat, 2),
            (self.user_contributor_cat, 2),
            (self.user_guest_cat, 2),
            (self.user_viewer_cat, 0),
            (self.user_finder_cat, 0),
            (self.user_owner, 2),
            (self.user_delegate, 2),
            (self.user_contributor, 2),
            (self.user_guest, 2),
            (self.user_viewer, 0),
            (self.user_no_roles, 0),
        ]
        url = self.url + '?' + urlencode({'s': 'link type:link'})
        self.assert_element_count(expected, url, 'sodar-ff-search-item')

    def test_search_type_nonexisting(self):
        """Test search items visibility with a nonexisting type"""
        expected = [
            (self.superuser, 0),
            (self.user_owner_cat, 0),
            (self.user_delegate_cat, 0),
            (self.user_contributor_cat, 0),
            (self.user_guest_cat, 0),
            (self.user_viewer_cat, 0),
            (self.user_finder_cat, 0),
            (self.user_owner, 0),
            (self.user_delegate, 0),
            (self.user_contributor, 0),
            (self.user_guest, 0),
            (self.user_viewer, 0),
            (self.user_no_roles, 0),
        ]
        url = self.url + '?' + urlencode({'s': 'test type:Jaix1au'})
        self.assert_element_count(expected, url, 'sodar-ff-search-item')


class TestHomeView(UITestBase):
    """Tests for appearance of filesfolders specific data in the home view"""

    def test_project_list(self):
        """Test custom filesfolders project list column visibility"""
        users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
            self.user_no_roles,
        ]
        url = reverse('home')
        self.assert_element_exists(
            users, url, 'sodar-pr-project-list-header-filesfolders-files', True
        )
        self.assert_element_exists(
            users, url, 'sodar-pr-project-list-header-filesfolders-links', True
        )



"""Forms for the userprofile app"""

from typing import Optional

from django import forms

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.forms import (
    SODARForm,
    SODARModelForm,
    SODARAppSettingFormMixin,
)
from projectroles.models import (
    SODARUser,
    SODARUserAdditionalEmail,
    SODAR_CONSTANTS,
    ADD_EMAIL_ALREADY_SET_MSG,
)
from projectroles.plugins import PluginAPI
from projectroles.utils import build_secret


app_settings = AppSettingAPI()
plugin_api = PluginAPI()


# SODAR Constants
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']


class UserAppSettingsForm(SODARAppSettingFormMixin, SODARForm):
    """Form for configuring user settings"""

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('current_user')
        super().__init__(*args, **kwargs)
        # Init app settings fields
        project_plugins = plugin_api.get_active_plugins(
            plugin_type='project_app'
        )
        site_plugins = plugin_api.get_active_plugins(plugin_type='site_app')
        self.app_plugins = project_plugins + site_plugins
        user_mod = False if self.user.is_superuser else True
        self.init_app_settings(
            self.app_plugins, APP_SETTING_SCOPE_USER, user_mod
        )

    def clean(self):
        cleaned_data, errors = self.clean_app_settings(
            self.cleaned_data,
            self.app_plugins,
            APP_SETTING_SCOPE_USER,
            False if self.user.is_superuser else True,
            user=self.user,
        )
        self.cleaned_data = cleaned_data
        for field, error in errors:
            self.add_error(field, error)
        return self.cleaned_data


class UserEmailForm(SODARModelForm):
    """Form for creating additional email addresses for user"""

    class Meta:
        model = SODARUserAdditionalEmail
        fields = ['email', 'user', 'secret']

    def __init__(
        self, current_user: Optional[SODARUser] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if current_user:
            self.current_user = current_user
        self.fields['user'].widget = forms.HiddenInput()
        self.initial['user'] = current_user
        self.fields['secret'].widget = forms.HiddenInput()
        self.initial['secret'] = build_secret(32)

    def clean(self):
        if self.cleaned_data['email'] == self.current_user.email:
            self.add_error(
                'email', ADD_EMAIL_ALREADY_SET_MSG.format(email_type='primary')
            )
            return self.cleaned_data
        if (
            SODARUserAdditionalEmail.objects.filter(
                user=self.current_user, email=self.cleaned_data['email']
            ).count()
            > 0
        ):
            self.add_error(
                'email',
                ADD_EMAIL_ALREADY_SET_MSG.format(email_type='additional'),
            )
        return self.cleaned_data



# Generated by Django 4.2.14 on 2024-07-16 09:22

from django.conf import settings
from django.db import migrations, models
import django.utils.timezone
import uuid


class Migration(migrations.Migration):

    replaces = [
        ('bgjobs', '0001_initial'),
        ('bgjobs', '0002_backgroundjoblogentry'),
        ('bgjobs', '0003_backgroundjob_date_created'),
        ('bgjobs', '0004_backgroundjob_user'),
        ('bgjobs', '0005_auto_20190128_1210'),
        ('bgjobs', '0006_auto_20200526_1657'),
    ]

    initial = True

    dependencies = [
        ('projectroles', '0005_update_uuid'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='BackgroundJob',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'date_modified',
                    models.DateTimeField(
                        auto_now=True, help_text='DateTime of last modification'
                    ),
                ),
                (
                    'sodar_uuid',
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text='BG Job SODAR UUID',
                        unique=True,
                    ),
                ),
                (
                    'job_type',
                    models.CharField(
                        help_text='Type of the job', max_length=512
                    ),
                ),
                ('name', models.CharField(max_length=512)),
                ('description', models.TextField()),
                (
                    'status',
                    models.CharField(
                        choices=[
                            ('initial', 'initial'),
                            ('running', 'running'),
                            ('done', 'done'),
                            ('failed', 'failed'),
                        ],
                        default='initial',
                        max_length=50,
                    ),
                ),
                (
                    'project',
                    models.ForeignKey(
                        help_text='Project in which this objects belongs',
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to='projectroles.project',
                    ),
                ),
                (
                    'date_created',
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text='DateTime of creation',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='background_jobs',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'ordering': ['-date_created'],
            },
        ),
        migrations.CreateModel(
            name='BackgroundJobLogEntry',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'date_created',
                    models.DateTimeField(
                        auto_now_add=True, help_text='DateTime of creation'
                    ),
                ),
                (
                    'level',
                    models.CharField(
                        choices=[
                            ('debug', 'debug'),
                            ('info', 'info'),
                            ('warning', 'warning'),
                            ('error', 'error'),
                        ],
                        help_text='Level of log entry',
                        max_length=50,
                    ),
                ),
                ('message', models.TextField(help_text='Log level\'s message')),
                (
                    'job',
                    models.ForeignKey(
                        help_text='Owning background job',
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='log_entries',
                        to='bgjobs.backgroundjob',
                    ),
                ),
            ],
            options={
                'ordering': ['date_created'],
            },
        ),
    ]
